import numpy as np
import pandas as pd
from astropy.table import Table
from citation_compass import cite_function

from lightcurvelynx import _LIGHTCURVELYNX_BASE_DATA_DIR
from lightcurvelynx.astro_utils.mag_flux import mag2flux
from lightcurvelynx.astro_utils.noise_model import poisson_bandflux_std
from lightcurvelynx.obstable.obs_table import ObsTable

ROMAN_PIXEL_SCALE = 0.11
"""The pixel scale for Roman in arcseconds per pixel.
https://roman-docs.stsci.edu/roman-instruments-home/wfi-imaging-mode-user-guide/introduction-to-the-wfi/wfi-quick-reference
"""

_roman_dark_current = 0.0
"""The dark current for Roman in electrons per second per pixel.
< 0.005 e-/s/pix
https://roman-docs.stsci.edu/roman-instruments-home/wfi-imaging-mode-user-guide/wfi-detectors
"""

_roman_fov = 0.28
"""The Roman field of view in square degree
https://roman-docs.stsci.edu/roman-instruments-home/wfi-imaging-mode-user-guide/introduction-to-the-wfi/wfi-quick-reference
"""

_roman_zodi_level_factor = 1.5
"""
Factor for calculating Zodiacal light intensity.

According to the WFI technical documentation:
https://roman.gsfc.nasa.gov/science/WFI_technical.html

- At high galactic latitudes, the Zodi intensity is typically ~1.5× the
  minimum.
- For observation into the galactic bulge, the Zodi intensity is typically 2.5-7x the minimum.
"""


_psf_url = "https://raw.githubusercontent.com/RomanSpaceTelescope/roman-technical-information/refs/heads/main/data/WideFieldInstrument/Imaging/PointSpreadFunctions/SummaryPSFstats_center.ecsv"
_zp_url = "https://raw.githubusercontent.com/RomanSpaceTelescope/roman-technical-information/refs/heads/main/data/WideFieldInstrument/Imaging/ZeroPoints/Roman_zeropoints_20240301.ecsv"
_thermal_url = "https://raw.githubusercontent.com/RomanSpaceTelescope/roman-technical-information/refs/heads/main/data/WideFieldInstrument/Imaging/Backgrounds/internal_thermal_backgrounds.ecsv"
_zodiacal_url = "https://raw.githubusercontent.com/RomanSpaceTelescope/roman-technical-information/refs/heads/main/data/WideFieldInstrument/Imaging/ZodiacalLight/zodiacal_light.ecsv"


def _get_roman_char():
    psf_table = Table.read(_psf_url, format="csv", comment="#").to_pandas()
    zp_table = Table.read(_zp_url, format="csv", comment="#", delimiter=" ").to_pandas()
    thermal_table = Table.read(_thermal_url, format="csv", comment="#").to_pandas()
    zodiacal_table = Table.read(_zodiacal_url, format="csv", comment="#").to_pandas()

    return {
        "psf_table": psf_table,
        "zp_table": zp_table,
        "thermal_table": thermal_table,
        "zodiacal_min_table": zodiacal_table,
    }


def _get_ma_table(ma_table_path=None):
    if ma_table_path is None:
        ma_table_file = (
            _LIGHTCURVELYNX_BASE_DATA_DIR
            / "roman_characterization/roman_wfi_imaging_multiaccum_tables_with_exptime.csv"
        )
    ma_table = pd.read_csv(ma_table_file)

    return ma_table


class RomanObsTable(ObsTable):
    """A subclass for Roman exposure table.

    Parameters
    ----------
    table : dict or pandas.core.frame.DataFrame
        The table with all the observation information.
    colmap : dict
        A mapping of short column names to their names in the underlying table.
        Defaults to the Roman APT column names, stored in _default_colnames.
    **kwargs : dict
        Additional keyword arguments to pass to the ObsTable constructor. This includes overrides
        for survey parameters such as:
        - dark_current : The dark current for the camera in electrons per second per pixel.
        - gain: The CCD gain (in e-/ADU).
        - pixel_scale: The pixel scale for the camera in arcseconds per pixel.
        - radius: The angular radius of the observations (in degrees).
        - read_noise: The standard deviation of the count of readout electrons per pixel.
    """

    # Default column names for the ZTF survey data.
    _default_colnames = {"dec": "DEC", "filter": "BANDPASS", "ra": "RA", "zp": "zp_nJy", "time": "time"}

    # Default survey values.
    _default_survey_values = {
        "dark_current": _roman_dark_current,
        "pixel_scale": ROMAN_PIXEL_SCALE,
        "survey_name": "Roman",
        "radius": np.sqrt(_roman_fov / np.pi),
        "zodi_level": _roman_zodi_level_factor,
    }

    def __init__(self, table, colmap=None, ma_table_path=None, **kwargs):
        colmap = self._default_colnames if colmap is None else colmap

        self.apt_table = table
        self.ma_table = _get_ma_table(ma_table_path)
        roman_char = _get_roman_char()
        self.zp_table = roman_char["zp_table"]
        self.psf_table = roman_char["psf_table"]
        self.thermal_table = roman_char["thermal_table"]
        self.zodiacal_table = roman_char["zodiacal_min_table"]
        self._append_apt_table()

        super().__init__(self.apt_table, colmap=colmap, **kwargs)

    def _append_apt_table(self):
        self.apt_table["zp_abmag"] = 0.0
        self.apt_table["N_Eff_Pix"] = 0.0
        self.apt_table["zodi_countrate_min"] = 0.0
        self.apt_table["thermal_countrate"] = 0.0
        self.apt_table["time"] = 0.0
        self.apt_table["exptime"] = 0.0

        for f in np.unique(self.apt_table.BANDPASS):
            zp_abmag = self.zp_table.loc[self.zp_table.element == f, "ABMag"].values[0]
            self.apt_table.loc[f == self.apt_table.BANDPASS, "zp_abmag"] = zp_abmag
            n_eff_pix = self.psf_table.loc[self.psf_table["Filter"] == f, "N_Eff_Pix"].values[0]
            self.apt_table.loc[f == self.apt_table.BANDPASS, "N_Eff_Pix"] = n_eff_pix
            sigma_zodi_min = self.zodiacal_table.loc[self.zodiacal_table["filter"] == f, "rate"].values[0]
            self.apt_table.loc[f == self.apt_table.BANDPASS, "zodi_countrate_min"] = sigma_zodi_min
            sigma_thermal = self.thermal_table.loc[self.thermal_table["filter"] == f, "rate"].values[0]
            self.apt_table.loc[f == self.apt_table.BANDPASS, "thermal_countrate"] = sigma_thermal

        for mat_number in np.unique(self.apt_table.MA_TABLE_NUMBER):
            exptime = self.ma_table.loc[self.ma_table["MATableNumber"] == mat_number, "Exptime"].values[0]
            self.apt_table.loc[mat_number == self.apt_table.MA_TABLE_NUMBER, "exptime"] = exptime

        npass = len(np.unique(self.apt_table.PASS))
        self.start_time = 61304.0
        self.cadence = 5.0
        times = self.start_time + self.cadence * np.arange(npass)
        for i, pass_num in enumerate(np.unique(self.apt_table.PASS)):
            self.apt_table.loc[pass_num == self.apt_table.PASS, "time"] = times[i]

        self.apt_table["zp_nJy"] = mag2flux(self.apt_table["zp_abmag"])

    @cite_function
    def readnoise_func(self, exptime):
        """
        Readout noise function for Roman.

        References
        ----------
        Eq. 9 of Rose et al 2025 - https://ui.adsabs.harvard.edu/abs/2025ApJ...988...65R/abstract

        Parameters
        ----------
        exptime: float or npt.ArrayLike
            Exposure time in seconds.

        Returns
        -------
        sigma_read: float or npt.ArrayLike
            Readout noise.
        """
        exptime = np.asarray(exptime)
        sigma_floor_sq = 25.0
        sigma_sq = 3072.0
        n = exptime / 3.04
        nfactor = (n - 1) / n / (n + 1)
        sigma_read = np.sqrt(sigma_floor_sq + sigma_sq * nfactor)

        return sigma_read

    @cite_function
    def calculate_skynoise(self, exptime, zodi_scale, zodi_countrate_min, thermal_countrate):
        """Calculate sky noise.

        Reference
        ---------
        Eq. 10 of Rose et al 2025 - https://ui.adsabs.harvard.edu/abs/2025ApJ...988...65R/abstract

        Parameters
        ----------
        exptime: float or npt.ArrayLike
            Exposure time.
        zodi_scale: float or npt.ArrayLike
            Zodiacal light scale. The zodiacal light amount is zodi_scale * zodi_countrate_min
        zodi_countrate_min: float or npt.ArrayLike
            Minimum zodiacal count rate (e-/s/pixel).
        thermal_countrate: float or npt.ArrayLike
            Thermal count rate (e-/s/pixel).

        Returns
        -------
        sky_variance: float or npt.ArrayLike
            Total sky variance (e-^2/pixel).
        """

        exptime = np.asarray(exptime)
        zodi_scale = np.asarray(zodi_scale)
        zodi_countrate_min = np.asarray(zodi_countrate_min)
        thermal_countrate = np.asarray(thermal_countrate)
        sky_variance = exptime * ((zodi_scale * zodi_countrate_min) + thermal_countrate)

        return sky_variance

    def bandflux_error_point_source(self, bandflux, index):
        """Compute observational bandflux error for a point source

        Parameters
        ----------
        bandflux : array_like of float
            Band bandflux of the point source in nJy.
        index : array_like of int
            The index of the observation in the table.

        Returns
        -------
        flux_err : array_like of float
            Simulated bandflux noise in nJy.
        """
        observations = self._table.iloc[index]
        observations["sky"] = self.calculate_skynoise(
            observations["exptime"],
            self.safe_get_survey_value("zodi_level"),
            observations["zodi_countrate_min"],
            observations["thermal_countrate"],
        )

        return poisson_bandflux_std(
            bandflux,  # nJy
            total_exposure_time=observations["exptime"],
            exposure_count=1,
            psf_footprint=observations["N_Eff_Pix"],
            sky=observations["sky"],
            zp=observations["zp"] / observations["exptime"],  # (nJy/s * s)^-1
            readout_noise=self.readnoise_func,  # e-/pixel
            dark_current=self.safe_get_survey_value("dark_current"),  # e-/second/pixel
        )



"""The SplineModel represents SED functions as a two dimensional grid
of (time, wavelength) -> flux value that is interpolated using a 2D spline.

It is adapted from sncosmo's TimeSeriesSource model:
https://github.com/sncosmo/sncosmo/blob/v2.10.1/sncosmo/models.py
"""

from scipy.interpolate import RectBivariateSpline

from lightcurvelynx.models.physical_model import SEDModel
from lightcurvelynx.utils.io_utils import read_grid_data


class SplineModel(SEDModel):
    """A time series model defined by sample points where the intermediate
    points are fit by a spline. Based on sncosmo's TimeSeriesSource:
    https://github.com/sncosmo/sncosmo/blob/v2.10.1/sncosmo/models.py

    The model is defined by a 2D grid of flux values as a function of time
    and wavelength. Time is given in decimal days since t0 and wavelength
    is given in angstroms.

    Parameterized values include:
      * amplitude - A unitless scaling parameter for the flux density values.
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - The time corresponding to the t=0.0 in the model data. [from BasePhysicalModel]

    Attributes
    ----------
    _times : numpy.ndarray
        A length T array containing the times relative to t0 at which the data
        was sampled.
    _wavelengths : numpy.ndarray
        A length W array containing the wavelengths at which the data was sampled
        (in angstroms).
    _spline : RectBivariateSpline
        The spline object for predicting the flux from a given (time, wavelength).

    Parameters
    ----------
    times : numpy.ndarray
        A length T array containing the times at which the data was sampled.
    wavelengths : numpy.ndarray
        A length W array containing the wavelengths at which the data was sampled
        (in angstroms).
    flux : numpy.ndarray
        A shape (T, W) matrix with flux values for each pair of time and wavelength.
        Fluxes provided in erg / s / cm^2 / Angstrom.
    amplitude : parameter
        A unitless scaling parameter for the flux density values. Default: 1.0
    time_degree : int
        The polynomial degree to use in the time dimension.
    wave_degree : int
        The polynomial degree to use in the wavelength dimension.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(
        self,
        times,
        wavelengths,
        flux,
        amplitude=1.0,
        time_degree=3,
        wave_degree=3,
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Set the attributes that can be changed (e.g. sampled)
        # t0 is set in the parent class.
        self.add_parameter("amplitude", amplitude, **kwargs)

        # These parameters are directly set, because they cannot be changed once
        # the object is created.
        self._times = times
        self._wavelengths = wavelengths
        self._spline = RectBivariateSpline(times, wavelengths, flux, kx=time_degree, ky=wave_degree)

    @classmethod
    def from_file(
        cls,
        input_file,
        format="ascii",
        amplitude=1.0,
        time_degree=3,
        wave_degree=3,
        **kwargs,
    ):
        """Create a SplineModel from a file where the data consists
        of three columns: time, wavelength, and the corresponding flux.

        The format of the data should be:
            time_0, wave_0, flux_0_0
            time_0, wave_1, flux_0_1
            ...
            time_0, wave_N, flux_0_N
            time_1, wave_0, flux_1_0
            time_1, wave_1, flux_1_1
            ...
            time_M, wave_N, flux_M_N
        The time and wavelength values must be sorted in ascending order.

        Parameters
        ----------
        input_file : str
            The input data file.
        format : str, optional
            The format of the input data such as 'ascii', 'ascii.ecsv', or 'fits'.
            Default: 'ascii'
        amplitude : parameter
            A unitless scaling parameter for the flux density values. Default: 1.0
        time_degree : int
            The polynomial degree to use in the time dimension.
        wave_degree : int
            The polynomial degree to use in the wavelength dimension.
        **kwargs : dict, optional
            Any additional keyword arguments.

        Returns
        -------
        SplineModel
            An instance of the SplineModel class.
        """
        times, wavelengths, flux = read_grid_data(input_file, format=format, validate=True)
        return cls(times, wavelengths, flux, amplitude, time_degree, wave_degree, **kwargs)

    def minwave(self, graph_state=None):
        """Get the minimum wavelength of the model.

        Parameters
        ----------
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. Not used
            for this model.

        Returns
        -------
        minwave : float or None
            The minimum wavelength of the model (in angstroms) or None
            if the model does not have a defined minimum wavelength.
        """
        return self._wavelengths[0]

    def maxwave(self, graph_state=None):
        """Get the maximum wavelength of the model.

        Parameters
        ----------
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. Not used
            for this model.

        Returns
        -------
        maximum : float or None
            The maximum wavelength of the model (in angstroms) or None
            if the model does not have a defined maximum wavelength.
        """
        return self._wavelengths[-1]

    def compute_sed(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
           Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        params = self.get_local_params(graph_state)
        t0 = params["t0"]
        if t0 is None:
            t0 = 0.0

        fluxes = params["amplitude"] * self._spline(times - t0, wavelengths, grid=True)
        return fluxes



import numpy as np
from astropy.cosmology import FlatLambdaCDM
from lightcurvelynx.astro_utils.snia_utils import num_snia_per_redshift_bin, snia_volumetric_rates


def test_snia_volumetric_rates():
    """
    Test that the volumetric rates are correctly calculated.
    """

    z = np.array([0.0, 1.0])
    expected_rates = [2.27e-5, 7.375e-5]
    rates = snia_volumetric_rates(z)

    assert np.allclose(rates, expected_rates)


def test_num_snia_per_redshift_bin():
    """
    Test that the number of SN Ia calculated matches what it would be in a small bin without integration.
    """

    H0 = 70.0
    Omega_m = 0.3

    # zmean = 0.073
    zmin = 0.07
    zmax = 0.08
    znbins = 1
    nsn, z = num_snia_per_redshift_bin(zmin, zmax, znbins, H0=H0, Omega_m=Omega_m)

    cosmo = FlatLambdaCDM(H0=H0, Om0=Omega_m)

    expected_rate = 2.42e-5
    expected_nsn = expected_rate * (cosmo.comoving_volume(zmax) - cosmo.comoving_volume(zmin))
    expected_nsn = expected_nsn.value

    nsn_total = np.sum(nsn)

    assert z.size == znbins
    assert nsn.size == znbins
    assert np.allclose(nsn_total, expected_nsn, rtol=0.1)



from unittest.mock import patch

import numpy as np
import pytest
from lightcurvelynx.astro_utils.passbands import Passband, PassbandGroup
from lightcurvelynx.models.spline_model import SplineModel
from sncosmo import Bandpass


def create_lsst_passband_group(passbands_dir, delta_wave=5.0, trim_quantile=None):
    """Helper function to create a PassbandGroup object for LSST passbands, using transmission tables
    included in the test data directory."""
    return PassbandGroup.from_preset(
        preset="LSST",
        table_dir=passbands_dir,
        delta_wave=delta_wave,
        trim_quantile=trim_quantile,
    )


def create_toy_passband_group(path, delta_wave=5.0, trim_quantile=None):
    """Helper function to create a toy PassbandGroup object for testing.

    Notable features of this toy passband group:
    - Three passbands: A, B, C
    - Each transmission table includes only three rows
    - Passbands A and B have overlapping wavelength ranges
    - Passbands A and C are defined with a 100 unit grid step, while passband B is defined with a 50 unit step
    """
    # Initialize requirements for PassbandGroup object
    survey = "TOY"
    filter_names = ["a", "b", "c"]
    table_dir = path / survey
    table_dir.mkdir()

    transmission_tables = {
        "a": "100 0.5\n200 0.75\n300 0.25\n",
        "b": "250 0.25\n300 0.5\n350 0.75\n",
        "c": "400 0.75\n500 0.25\n600 0.5\n",
    }
    for filter_name in filter_names:
        with open(table_dir / f"{filter_name}.dat", "w") as f:
            f.write(transmission_tables[filter_name])

    # Load the PassbandGroup object
    passbands = []
    for filter_name in filter_names:
        passbands.append(
            {
                "survey": survey,
                "filter_name": filter_name,
                "table_path": table_dir / f"{filter_name}.dat",
                "delta_wave": delta_wave,
                "trim_quantile": trim_quantile,
            }
        )
    return PassbandGroup(passbands)


def test_passband_group_access(tmp_path):
    """Test that we can create a passband group and access the individual passbands."""
    table_vals = np.array([[100, 0.5], [200, 0.75], [300, 0.25]])
    pb_list = [
        Passband(table_vals, "survey1", "a", trim_quantile=None),
        Passband(table_vals, "survey1", "b", trim_quantile=None),
        Passband(table_vals, "survey1", "c", trim_quantile=None),
        Passband(table_vals, "survey2", "c", trim_quantile=None),
        Passband(table_vals, "survey2", "d", trim_quantile=None),
    ]

    pb_group = PassbandGroup(given_passbands=pb_list)
    assert len(pb_group) == 5
    assert "survey1_a" in pb_group
    assert "survey1_b" in pb_group
    assert "survey1_c" in pb_group
    assert "survey2_c" in pb_group
    assert "survey2_d" in pb_group

    # We can access the passbands by their full name.
    assert pb_group["survey1_a"].filter_name == "a"
    assert pb_group["survey1_b"].filter_name == "b"
    assert pb_group["survey1_c"].filter_name == "c"
    assert pb_group["survey2_c"].filter_name == "c"
    assert pb_group["survey2_d"].filter_name == "d"

    # Test that we can access a list of filters.
    assert pb_group.filters == ["a", "b", "c", "d"]

    # If the group only has a single passband for a given filter,
    # We can use the filter name to access it.
    assert pb_group["a"].full_name == "survey1_a"
    assert pb_group["b"].full_name == "survey1_b"
    assert pb_group["d"].full_name == "survey2_d"

    # But we get an error if we try to look up by a filter name that occurs twice.
    with pytest.raises(KeyError):
        _ = pb_group["c"]

    # The contains functionality will work with all filter names.
    assert "a" in pb_group
    assert "b" in pb_group
    assert "c" in pb_group
    assert "d" in pb_group

    # Check that we can iterate through the passbands.
    for idx, pb in enumerate(pb_group):
        assert pb == pb_list[idx]

    # Check that we can filter a list of filter_names by whether they occur in the passband group.
    filters = ["a", "b", "e", "f", "c", "1", "2", "d", "a", "a"]
    expected = [True, True, False, False, True, False, False, True, True, True]
    assert np.array_equal(pb_group.mask_by_filter(filters), expected)


def test_passband_group_init(tmp_path, passbands_dir):
    """Test the initialization of the Passband class, and implicitly, _load_preset."""
    # Test that we cannot create an empty PassbandGroup object
    with pytest.raises(TypeError):
        _ = PassbandGroup()

    # Test that the PassbandGroup class raises an error for an unknown preset
    try:
        _ = PassbandGroup.from_preset(preset="Unknown")
    except ValueError as e:
        assert str(e) == "Unknown passband preset: unknown"
    else:
        raise AssertionError("PassbandGroup should raise an error for an unknown preset")

    # Test that the PassbandGroup class can be initialized with a preset
    lsst_passband_group = create_lsst_passband_group(passbands_dir)
    assert len(lsst_passband_group.passbands) == 6
    assert len(lsst_passband_group) == 6
    assert "LSST_u" in lsst_passband_group
    assert "LSST_g" in lsst_passband_group
    assert "LSST_r" in lsst_passband_group
    assert "LSST_i" in lsst_passband_group
    assert "LSST_z" in lsst_passband_group
    assert "LSST_y" in lsst_passband_group
    assert "LSST_purple" not in lsst_passband_group

    # Test that we fail if we use the wrong units for the LSST preset
    with pytest.raises(ValueError):
        _ = PassbandGroup.from_preset(
            preset="LSST",
            table_dir=passbands_dir,
            delta_wave=5.0,
            trim_quantile=None,
            units="A",
        )

    # We can access passbands using the [] notation.
    assert lsst_passband_group["LSST_u"].filter_name == "u"
    with pytest.raises(KeyError):
        _ = lsst_passband_group["LSST_purple"]

    # Test that the PassbandGroup class can be initialized with a dict of passband parameters
    lsst_gri_passband_parameters = [
        {"survey": "LSST", "filter_name": "g", "table_path": f"{passbands_dir}/LSST/g.dat"},
        {"survey": "LSST", "filter_name": "r", "table_path": f"{passbands_dir}/LSST/r.dat"},
        {"survey": "LSST", "filter_name": "i", "table_path": f"{passbands_dir}/LSST/i.dat"},
    ]
    lsst_gri_passband_group = PassbandGroup(given_passbands=lsst_gri_passband_parameters)
    assert len(lsst_gri_passband_group) == 3
    assert "LSST_g" in lsst_gri_passband_group
    assert "LSST_r" in lsst_gri_passband_group
    assert "LSST_i" in lsst_gri_passband_group

    # Test our toy passband group, which makes a PassbandGroup using a custom passband parameters dictionary
    toy_passband_group = create_toy_passband_group(tmp_path)
    assert len(toy_passband_group) == 3
    assert "TOY_a" in toy_passband_group
    assert "TOY_b" in toy_passband_group
    assert "TOY_c" in toy_passband_group
    np.testing.assert_allclose(
        toy_passband_group.waves,
        np.unique(np.concatenate([np.arange(100, 301, 5), np.arange(250, 351, 5), np.arange(400, 601, 5)])),
    )

    # Test that we can retrieve bounds for this passband group.
    min_w, max_w = toy_passband_group.wave_bounds()
    assert min_w == 100.0
    assert max_w == 600.0

    # Test that the PassbandGroup class raises an error for an unknown preset
    try:
        _ = PassbandGroup.from_preset(preset="Unknown")
    except ValueError as e:
        assert str(e) == "Unknown passband preset: unknown"
    else:
        raise AssertionError("PassbandGroup should raise an error for an unknown preset")


def test_passband_group_ztf_preset(passbands_dir):
    """Test that we can load a PassbandGroup using the ZTF preset."""
    pbg = PassbandGroup.from_preset(
        preset="ZTF",
        table_dir=passbands_dir,
        delta_wave=5.0,
        trim_quantile=None,
    )

    # Test that all of the filters are there
    assert len(pbg.passbands) == 3
    assert len(pbg) == 3
    assert "g" in pbg
    assert "r" in pbg
    assert "i" in pbg


def test_passband_group_roman_preset(passbands_dir):
    """Test that we can load a PassbandGroup using the Roman preset."""
    # Load the cached (older and simplified) Roman passband files from the test data directory.
    pbg = PassbandGroup.from_preset(
        preset="Roman",
        table_dir=passbands_dir,
        delta_wave=5.0,
        trim_quantile=None,
    )

    # Test that all of the filters are there
    assert len(pbg.passbands) == 8
    assert len(pbg) == 8
    assert "F062" in pbg
    assert "F087" in pbg
    assert "F106" in pbg
    assert "F129" in pbg
    assert "F158" in pbg
    assert "F184" in pbg
    assert "F146" in pbg
    assert "F213" in pbg

    # We can also load a subset of the filters.
    pbg_subset = PassbandGroup.from_preset(
        preset="Roman",
        table_dir=passbands_dir,
        delta_wave=5.0,
        trim_quantile=None,
        filters=["F062", "F106", "F158"],
    )

    # Test that the subset contains the correct filters.
    assert len(pbg_subset.passbands) == 3
    assert "F062" in pbg_subset
    assert "F106" in pbg_subset
    assert "F158" in pbg_subset
    assert "F087" not in pbg_subset
    assert "F129" not in pbg_subset
    assert "F184" not in pbg_subset
    assert "F146" not in pbg_subset
    assert "F213" not in pbg_subset

    # The load fails if we given the wrong units.
    with pytest.raises(ValueError):
        _ = PassbandGroup.from_preset(
            preset="Roman",
            table_dir=passbands_dir,
            delta_wave=5.0,
            trim_quantile=None,
            units="A",
            filters=["F062", "F106", "F158", "F999"],
        )


def test_passband_group_from_dir(tmp_path):
    """Test that we can load a PassbandGroup from a directory."""
    survey = "FAKE"

    # Create a new survey directory and fill it with filter files.
    table_dir = tmp_path / survey
    table_dir.mkdir()
    transmission_tables = {
        "a": "100 0.5\n150 0.75\n200 0.25\n",
        "b": "200 0.25\n250 0.5\n300 0.75\n",
        "c": "300 0.75\n350 0.25\n400 0.5\n",
        "d": "400 0.75\n450 0.25\n500 0.5\n",
        "e": "500 0.75\n550 0.25\n600 0.5\n",
    }
    for filter_name in transmission_tables:
        with open(table_dir / f"{filter_name}.dat", "w") as f:
            f.write(transmission_tables[filter_name])

    # Load four of the filters from the PassbandGroup
    load_filters = ["a", "b", "c", "e"]
    pb_group = PassbandGroup.from_dir(table_dir, filters=load_filters)
    assert len(pb_group) == len(load_filters)

    for filter in load_filters:
        assert filter in pb_group

        wave_start = int(transmission_tables[filter].split()[0])
        assert wave_start == pb_group[filter]._loaded_table[0][0]

    # Check that we throw an error if we try to access an invalid directory.
    with pytest.raises(ValueError):
        _ = PassbandGroup.from_dir("./no_such_directory", filters=["a", "b"])

    # Check that we throw an error if we try to load a filter that does not exist.
    with pytest.raises(ValueError):
        _ = PassbandGroup.from_dir(table_dir, filters=["a", "b", "z"])


def test_passband_group_from_list(tmp_path):
    """Test that we can create a PassbandGroup from a pre-specified list."""
    pb_list = [
        Passband(
            np.array([[100, 0.5], [200, 0.75], [300, 0.25]]),
            "my_survey",
            "a",
            trim_quantile=None,
        ),
        Passband(
            np.array([[250, 0.25], [300, 0.5], [350, 0.75]]),
            "my_survey",
            "b",
            trim_quantile=None,
        ),
        Passband(
            np.array([[400, 0.75], [500, 0.25], [600, 0.5]]),
            "my_survey",
            "c",
            trim_quantile=None,
        ),
    ]
    test_passband_group = PassbandGroup(given_passbands=pb_list)
    assert len(test_passband_group) == 3
    assert "my_survey_a" in test_passband_group
    assert "my_survey_b" in test_passband_group
    assert "my_survey_c" in test_passband_group

    all_waves = np.concatenate([np.arange(100, 301, 5), np.arange(250, 351, 5), np.arange(400, 601, 5)])
    assert np.allclose(test_passband_group.waves, np.unique(all_waves))

    # Test that we can manually add another passband to the group.
    pb4 = Passband(
        np.array([[800, 0.75], [850, 0.25], [900, 0.5]]),
        "my_survey2",
        "a",
        trim_quantile=None,
    )
    test_passband_group.add_passband(pb4)

    assert len(test_passband_group) == 4
    assert "my_survey_a" in test_passband_group
    assert "my_survey_b" in test_passband_group
    assert "my_survey_c" in test_passband_group
    assert "my_survey2_a" in test_passband_group

    all_waves = np.concatenate([all_waves, np.arange(800, 901, 5)])
    assert np.allclose(test_passband_group.waves, np.unique(all_waves))


def test_passband_load_subset_passbands(tmp_path):
    """Test that we can load a subset of filters."""
    pb_list = [
        Passband(
            np.array([[100, 0.5], [200, 0.75], [300, 0.25]]),
            "my_survey",
            "a",
            trim_quantile=None,
        ),
        Passband(
            np.array([[250, 0.25], [300, 0.5], [350, 0.75]]),
            "my_survey",
            "b",
            trim_quantile=None,
        ),
        Passband(
            np.array([[400, 0.75], [500, 0.25], [600, 0.5]]),
            "my_survey",
            "c",
            trim_quantile=None,
        ),
        Passband(
            np.array([[800, 0.75], [850, 0.25], [900, 0.5]]),
            "my_survey",
            "d",
            trim_quantile=None,
        ),
    ]

    # Load one filter by full name and the other by filter name.
    test_passband_group = PassbandGroup(given_passbands=pb_list, filters=["my_survey_a", "c"])
    assert len(test_passband_group) == 2

    assert np.allclose(
        test_passband_group.waves,
        np.unique(np.concatenate([np.arange(100, 301, 5), np.arange(400, 601, 5)])),
    )

    # We run into an error if we try to load a filter that does not exist.
    with pytest.raises(ValueError):
        _ = PassbandGroup(given_passbands=pb_list, filters=["my_survey_a", "z"])


def test_passband_ztf_preset():
    """Test that we can load the ZTF passbands."""

    def _mock_get_bandpass(name):
        """Return a predefined Bandpass object instead of downloading the transmission table."""
        return Bandpass(np.array([6000, 6005, 6010]), np.array([0.5, 0.6, 0.7]))

    # Mock the get_bandpass portion of the download method
    with patch("sncosmo.get_bandpass", side_effect=_mock_get_bandpass):
        group = PassbandGroup.from_preset(preset="ZTF")
        assert len(group) == 3
        assert "ZTF_g" in group
        assert "ZTF_r" in group
        assert "ZTF_i" in group

    # Try the load with a subset of filters.
    with patch("sncosmo.get_bandpass", side_effect=_mock_get_bandpass):
        group = PassbandGroup.from_preset(preset="ZTF", filters=["g", "i"])
        assert len(group) == 2
        assert "ZTF_g" in group
        assert "ZTF_i" in group
        assert "ZTF_r" not in group


def test_passband_group_from_svo():
    """Test that we can load a PassbandGroup from the SVO database."""

    # Mock the Passband.from_svo method to return a predefined Passband object.
    # This function is already tested in test_passband.py.
    def _mock_from_svo(full_filter_name, *args, **kwargs):
        """Return a predefined Passband object instead of downloading the transmission table."""
        survey, filter = full_filter_name.split(".")
        return Passband(
            np.array([[6000, 0.5], [6005, 0.6], [6010, 0.7]]),
            survey=survey,
            filter_name=filter,
        )

    # Mock the get_bandpass portion of the download method
    with patch("lightcurvelynx.astro_utils.passbands.Passband.from_svo", side_effect=_mock_from_svo):
        group = PassbandGroup.from_svo(
            ["SLOAN/SDSS.u", "SLOAN/SDSS.g", "SLOAN/SDSS.r", "SLOAN/SDSS.i", "SLOAN/SDSS.z"]
        )
        assert len(group) == 5
        assert "SLOAN/SDSS_u" in group
        assert "SLOAN/SDSS_g" in group
        assert "SLOAN/SDSS_r" in group
        assert "SLOAN/SDSS_i" in group
        assert "SLOAN/SDSS_z" in group


def test_passband_invalid_preset():
    """Test that we throw an error when given an invalid preset name."""
    with pytest.raises(ValueError):
        _ = PassbandGroup.from_preset(preset="Invalid")


def test_passband_unique_waves():
    """Test that if we create two passbands with very similar wavelengths, they get merged."""
    pb_list = [
        Passband(
            np.array([[100, 0.5], [250, 0.25]]),
            "my_survey",
            "a",
            trim_quantile=None,
        ),
        Passband(
            np.array([[200.000001, 0.25], [300.000001, 0.5]]),
            "my_survey",
            "b",
            trim_quantile=None,
        ),
    ]
    test_passband_group = PassbandGroup(given_passbands=pb_list)
    assert np.allclose(
        test_passband_group.waves,
        np.concatenate([np.arange(100, 201, 5), np.arange(205.000001, 301, 5)]),
    )

    # Larger gaps won't register.
    pb_list = [
        Passband(
            np.array([[100, 0.5], [250, 0.25]]),
            "my_survey",
            "a",
            trim_quantile=None,
        ),
        Passband(
            np.array([[200.5, 0.25], [300.5, 0.5]]),
            "my_survey",
            "b",
            trim_quantile=None,
        ),
    ]
    test_passband_group = PassbandGroup(given_passbands=pb_list)
    assert np.allclose(
        test_passband_group.waves,
        np.unique(np.concatenate([np.arange(100, 251, 5), np.arange(200.5, 301, 5)])),
    )


def test_passband_group_str(passbands_dir, tmp_path):
    """Test the __str__ method of the PassbandGroup class."""
    toy_passband_group = create_toy_passband_group(tmp_path)
    assert str(toy_passband_group) == "PassbandGroup containing 3 passbands: TOY_a, TOY_b, TOY_c"

    lsst_passband_group = create_lsst_passband_group(passbands_dir)
    assert str(lsst_passband_group) == (
        "PassbandGroup containing 6 passbands: LSST_u, LSST_g, LSST_r, LSST_i, LSST_z, LSST_y"
    )


def test_passband_group_fluxes_to_bandfluxes(passbands_dir):
    """Test the fluxes_to_bandfluxes method of the PassbandGroup class."""
    # Test with simple passband group
    lsst_passband_group = create_lsst_passband_group(passbands_dir, delta_wave=20, trim_quantile=None)

    # Create some mock flux data with the same number of columns as wavelengths in our PassbandGroup
    flux = np.linspace(lsst_passband_group.waves * 10, 100, 5)
    bandfluxes = lsst_passband_group.fluxes_to_bandfluxes(flux)

    # Check structure of result is as expected (see Passband tests for value tests)
    assert len(bandfluxes) == 6
    for band_name in lsst_passband_group.passbands:
        assert band_name in bandfluxes
        assert bandfluxes[band_name].shape == (5,)

        # Check that we get the same result with fluxes_to_bandflux.
        bandflux = lsst_passband_group.fluxes_to_bandflux(flux, band_name)
        np.testing.assert_allclose(bandflux, bandfluxes[band_name])


def test_passband_group_wrapped_from_physical_source(passbands_dir, tmp_path):
    """Test evaluate_bandflux, SEDModel's wrapped version of PassbandGroup's fluxes_to_bandfluxes."""
    times = np.array([1.0, 2.0, 3.0])
    wavelengths = np.array([10.0, 20.0, 30.0])
    fluxes = np.array([[1.0, 5.0, 1.0], [5.0, 10.0, 5.0], [1.0, 5.0, 3.0]])
    model = SplineModel(times, wavelengths, fluxes, time_degree=1, wave_degree=1)
    state = model.sample_parameters()

    test_times = np.array([1.0, 1.5, 2.0, 2.5, 3.0, 3.5])

    # Test wrapper with a PassbandGroup as input (see Passband tests for single-band tests)
    # Using LSST passband group:
    lsst_passband_group = create_lsst_passband_group(passbands_dir, delta_wave=20, trim_quantile=None)
    n_lsst_bands = len(lsst_passband_group.passbands)
    n_times = len(test_times)

    fluxes_source_model = model.evaluate_bandfluxes(
        lsst_passband_group,
        times=np.repeat(test_times, n_lsst_bands),
        filters=np.tile(list(lsst_passband_group.passbands.keys()), n_times),
        state=state,
    )
    result_from_source_model = {
        filter_name: fluxes_source_model[i::n_lsst_bands]
        for i, filter_name in enumerate(lsst_passband_group.passbands)
    }
    evaluated_fluxes = model.evaluate_sed(test_times, lsst_passband_group.waves, state)
    result_from_passband_group = lsst_passband_group.fluxes_to_bandfluxes(evaluated_fluxes)

    # Check the two dicts are the same
    assert result_from_source_model.keys() == result_from_passband_group.keys()
    for key in result_from_source_model:
        np.testing.assert_allclose(result_from_source_model[key], result_from_passband_group[key])

    # Using toy passband group:
    toy_passband_group = create_toy_passband_group(tmp_path, delta_wave=20, trim_quantile=None)
    n_toy_bands = len(toy_passband_group.passbands)
    fluxes_source_model = model.evaluate_bandfluxes(
        toy_passband_group,
        times=np.repeat(test_times, n_toy_bands),
        filters=np.tile(list(toy_passband_group.passbands.keys()), n_times),
        state=state,
    )
    result_from_source_model = {
        filter_name: fluxes_source_model[i::n_toy_bands]
        for i, filter_name in enumerate(toy_passband_group.passbands)
    }
    evaluated_fluxes = model.evaluate_sed(test_times, toy_passband_group.waves, state)
    result_from_passband_group = toy_passband_group.fluxes_to_bandfluxes(evaluated_fluxes)

    # Check the two dicts are the same
    assert result_from_source_model.keys() == result_from_passband_group.keys()
    for key in result_from_source_model:
        np.testing.assert_allclose(result_from_source_model[key], result_from_passband_group[key])


def test_passband_group_calculate_in_band_wave_indices(passbands_dir, tmp_path):
    """Test the calculate_in_band_wave_indices method of the PassbandGroup class using both toy and LSST
    transmission tables."""
    # Using LSST passband group:
    lsst_passband_group = create_lsst_passband_group(passbands_dir, delta_wave=20, trim_quantile=None)

    # Make sure group.waves contains the union of all passband waves with no duplicates or other values
    np.testing.assert_allclose(
        lsst_passband_group.waves,
        np.unique(np.concatenate([passband.waves for passband in lsst_passband_group.passbands.values()])),
    )

    # Using toy passband group:
    toy_passband_group = create_toy_passband_group(tmp_path, delta_wave=20, trim_quantile=None)

    passband_A = toy_passband_group.passbands["TOY_a"]
    passband_B = toy_passband_group.passbands["TOY_b"]
    passband_C = toy_passband_group.passbands["TOY_c"]

    # Note that passband_A and passband_B have overlapping wavelength ranges
    # Where passband_A covers 100-300 and passband_B covers 250-350 (and passband_C covers 400-600)
    toy_a_inds = toy_passband_group._in_band_wave_indices["TOY_a"]
    np.testing.assert_allclose(toy_a_inds, np.array([0, 1, 2, 3, 4, 5, 6, 7, 9, 11, 13]))
    np.testing.assert_allclose(passband_A.waves, toy_passband_group.waves[toy_a_inds])

    toy_b_inds = toy_passband_group._in_band_wave_indices["TOY_b"]
    np.testing.assert_allclose(toy_b_inds, np.array([8, 10, 12, 14, 15, 16]))
    np.testing.assert_allclose(passband_B.waves, toy_passband_group.waves[toy_b_inds])

    toy_c_inds = toy_passband_group._in_band_wave_indices["TOY_c"]
    assert toy_c_inds == slice(17, 28)
    np.testing.assert_allclose(passband_C.waves, toy_passband_group.waves[toy_c_inds])


def test_write_and_read_passband_group(passbands_dir, tmp_path):
    """Test that we can write a PassbandGroup to a file and read it back in."""
    file_path = tmp_path / "test_passband_group.csv"

    # We can't read the file if it does not exist.
    with pytest.raises(FileNotFoundError):
        _ = PassbandGroup.from_preprocessed_file(file_path)

    pb_group = create_lsst_passband_group(passbands_dir)
    pb_group.to_file(file_path)
    assert file_path.is_file()

    # We can read the file back in.
    pb_group2 = PassbandGroup.from_preprocessed_file(file_path)
    assert pb_group2 is not None
    for key in pb_group.passbands:
        assert key in pb_group2.passbands
        assert pb_group.passbands[key] == pb_group2.passbands[key]

    # We get an error if we try to overwrite an existing file without permission.
    with pytest.raises(FileExistsError):
        pb_group.to_file(file_path, overwrite=False)



import numpy as np
import pytest
from lightcurvelynx.effects.white_noise import WhiteNoise
from lightcurvelynx.models.basic_models import ConstantSEDModel


def test_white_noise() -> None:
    """Test that we can sample and create a WhiteNoise object."""
    white_noise = WhiteNoise(white_noise_sigma=0.1)
    assert str(white_noise) == "WhiteNoise"
    assert repr(white_noise) == "WhiteNoise(white_noise_sigma)"

    # We can apply the noise.
    values = np.full((5, 3), 100.0)
    values = white_noise.apply(values, white_noise_sigma=0.1)
    assert not np.all(values == 100.0)
    assert np.all(np.abs(values - 100.0) <= 1.0)

    # We can override the default value using the parameters.
    values = white_noise.apply(values, white_noise_sigma=20.0)
    assert not np.all(values == 100.0)
    assert not np.all(np.abs(values - 100.0) <= 1.0)

    # We fail if we do not pass in the expected parameters.
    with pytest.raises(ValueError):
        _ = white_noise.apply(values)


def test_white_noise_bandflux() -> None:
    """Test that we can sample a WhiteNoise object at the bandflux level."""
    values = np.full(10, 100.0)

    # We can apply the noise.
    white_noise = WhiteNoise(white_noise_sigma=0.1)
    values = white_noise.apply_bandflux(values, white_noise_sigma=0.1)
    assert not np.all(values == 100.0)
    assert np.all(np.abs(values - 100.0) <= 1.0)

    # We can override the default value using the parameters.
    values = white_noise.apply_bandflux(values, white_noise_sigma=20.0)
    assert not np.all(values == 100.0)
    assert not np.all(np.abs(values - 100.0) <= 1.0)

    # We fail if we do not pass in the expected parameters.
    with pytest.raises(ValueError):
        _ = white_noise.apply_bandflux(values)


def test_constant_sed_model_white_noise() -> None:
    """Test that we can sample and create a ConstantSEDModel object with white noise."""
    model = ConstantSEDModel(
        brightness=10.0,
        node_label="my_constant_sed_model",
        seed=100,
    )
    assert len(model.rest_frame_effects) == 0
    assert len(model.obs_frame_effects) == 0

    # We can add the white noise effect.
    white_noise = WhiteNoise(white_noise_sigma=0.1)
    model.add_effect(white_noise)
    assert len(model.rest_frame_effects) == 1
    assert len(model.obs_frame_effects) == 0

    state = model.sample_parameters()
    times = np.array([1, 2, 3, 4, 5, 10])
    wavelengths = np.array([100.0, 200.0, 300.0])
    values = model.evaluate_sed(times, wavelengths, state)
    assert values.shape == (6, 3)

    # We get noisy values around 10.0.
    assert len(np.unique(values)) > 10
    assert np.all(np.abs(values - 10.0) < 3.0)

    # Test that if we pass in an rng, we control the randomness.
    values1 = model.evaluate_sed(times, wavelengths, state, rng_info=np.random.default_rng(100))
    values2 = model.evaluate_sed(times, wavelengths, state, rng_info=np.random.default_rng(100))
    assert not np.any(values1 == 10.0)
    assert np.all(values1 == values2)


def test_constant_sed_model_white_noise_obs_frame() -> None:
    """Test that we can make the WhiteNoise an observer frame effect.
    While this does not make physical sense, it allows us to test that code path.
    """
    model = ConstantSEDModel(
        brightness=10.0,
        node_label="my_constant_sed_model",
        seed=100,
    )
    assert len(model.rest_frame_effects) == 0
    assert len(model.obs_frame_effects) == 0

    white_noise = WhiteNoise(rest_frame=False, white_noise_sigma=0.1)
    model.add_effect(white_noise)
    assert len(model.rest_frame_effects) == 0
    assert len(model.obs_frame_effects) == 1

    state = model.sample_parameters()
    times = np.array([1, 2, 3, 4, 5, 10])
    wavelengths = np.array([100.0, 200.0, 300.0])
    values = model.evaluate_sed(times, wavelengths, state)
    assert values.shape == (6, 3)

    # We get noisy values around 10.0.
    assert len(np.unique(values)) > 10
    assert np.all(np.abs(values - 10.0) < 3.0)



from lightcurvelynx import consts
from numpy.testing import assert_allclose


def test_parsec_to_cm():
    """Check PARSEC_TO_CM constant"""
    assert_allclose(consts.PARSEC_TO_CM, 3.08567758149137e18)


def test_gauss_eff_area2fwhm_sq():
    """Test GAUSS_EFF_AREA2FWHM_SQ to be close to 2.266

    Find 2.266 here:
    https://smtn-002.lsst.io/v/OPSIM-1171/index.html
    """
    assert_allclose(consts.GAUSS_EFF_AREA2FWHM_SQ, 2.266, atol=1e-3)


def test_angstrom_to_cm():
    """Check ANGSTROM_TO_CM constant"""
    assert_allclose(consts.ANGSTROM_TO_CM, 1e-8)


def test_cgs_fnu_unit_to_njy():
    """Check CGS_FNU_UNIT_TO_NJY constant"""
    assert_allclose(consts.CGS_FNU_UNIT_TO_NJY, 1e32)



"""Wrappers for the models defined in redback.

https://github.com/nikhil-sarin/redback
https://redback.readthedocs.io/en/latest/
"""

import math

import astropy.units as uu
import numpy as np
from citation_compass import CiteClass, cite_inline

from lightcurvelynx.astro_utils.unit_utils import flam_to_fnu
from lightcurvelynx.math_nodes.bilby_priors import BilbyPriorNode
from lightcurvelynx.models.physical_model import SEDModel


class RedbackWrapperModel(SEDModel, CiteClass):
    """A wrapper for redback models.

    Parameterized values include:
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - The t0 of the zero phase, date. [from BasePhysicalModel]
    Additional parameterized values are used for specific redback models.

    References
    ----------
    * redback - https://ui.adsabs.harvard.edu/abs/2024MNRAS.531.1203S/abstract
    * Individual models might require citation. See references in the redback documentation.

    Attributes
    ----------
    source : function
        The underlying source function that maps time + wavelength to flux.
    source_name : str
        The name used to set the source.
    source_param_names : list
        A list of the source model's parameters that we need to set.

    Parameters
    ----------
    source : str or function
        The name of the redback model function used to generate the SEDs or
        the actual function itself.
    priors : dict, bilby.prior.PriorDict, or BilbyPriorNode, optional
        The redback model's Bilby priors.
    parameters : dict, optional
        A dictionary of parameter setters to pass to the source function.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    # A class variable for the units so we are not computing them each time.
    _FLAM_UNIT = uu.erg / uu.second / uu.cm**2 / uu.AA

    def __init__(
        self,
        source,
        *,
        priors=None,
        parameters=None,
        **kwargs,
    ):
        # Check that the parameters passed in the dictionary and keyword arguments
        # do not overlap, so we only have one source of truth. This is needed for
        # parameters like `redshift` that overlap core parameters.
        if parameters is None:
            parameters = {}
        for key in parameters:
            if key in kwargs:
                raise ValueError(
                    f"Parameter '{key}' specified in both the parameters dictionary "
                    "and as a parameter itself. Please include it only in the dictionary."
                )

        super().__init__(**kwargs)

        # Add all of the items from the bilby prior node as settable parameters.
        parameters = parameters.copy()
        if priors is not None:
            if not isinstance(priors, BilbyPriorNode):
                priors = BilbyPriorNode(prior=priors)
            for param_name in priors.outputs:
                if param_name not in parameters:
                    parameters[param_name] = getattr(priors, param_name)

        # Use the parameter dictionary to create settable parameters for the model.
        # Some of these might have already been added by the superclass's constructor,
        # so we just change it.
        self.source_param_names = []
        for key, value in parameters.items():
            if key in self.setters:
                self.set_parameter(key, value)
            else:
                self.add_parameter(key, value)
            self.source_param_names.append(key)

        # Create the source itself.
        if isinstance(source, str):
            try:
                import redback
            except ImportError as err:
                raise ImportError(
                    "redback package is not installed be default. To use the RedbackWrapperModel, "
                    "please install redback. For example, you can install it with "
                    "`pip install redback`."
                ) from err

            self.source_name = source
            self.source = redback.model_library.all_models_dict[source]
        else:
            self.source_name = source.__name__
            self.source = source

        # Check if the model has a citation parameter we should include.
        if hasattr(self.source, "citation"):
            cite_inline("redback model", self.source.citation)

        # Redback models already handle redshift, so we do not want to double apply it.
        self.apply_redshift = False

    @property
    def param_names(self):
        """Return a list of the model's parameter names."""
        return self.source_param_names

    def minwave(self, graph_state=None):
        """Get the minimum wavelength of the model.

        Parameters
        ----------
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. Not used
            for this model.

        Returns
        -------
        minwave : float or None
            The minimum wavelength of the model (in angstroms) or None
            if the model does not have a defined minimum wavelength.
        """
        return 0.0

    def maxwave(self, graph_state=None):
        """Get the maximum wavelength of the model.

        Parameters
        ----------
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. Not used
            for this model.

        Returns
        -------
        maxwave : float or None
            The maximum wavelength of the model (in angstroms) or None
            if the model does not have a defined maximum wavelength.
        """
        return math.inf

    def compute_sed(self, times, wavelengths, graph_state=None, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps (MJD).
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
           Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        params = self.get_local_params(graph_state)

        # Build the function arguments from the parameter values.
        fn_args = {}
        for name in self.source_param_names:
            fn_args[name] = params[name]

        # Compute the shifted times.
        t0 = params.get("t0", 0.0)
        if t0 is None:
            t0 = 0.0
        shifted_times = times - t0

        # Call the source function to get the RedbackTimeSeriesSource object.
        # We create this object with each call, because it depends on the parameters (fn_args).
        rb_result = self.source(
            shifted_times,
            output_format="sncosmo_source",
            **fn_args,
        )

        # The sncosmo-type source gives an error if the wavelengths are out of bounds, so we
        # need to use extrapolation if the wavelengths are out of bounds. The function
        # compute_sed_with_extrapolation() will trim the input wavelengths to be within the model's bounds
        # and call this function again. So the second call passes this bounds check and does not recurse.
        if np.any(wavelengths < rb_result.minwave()) or np.any(wavelengths > rb_result.maxwave()):
            return self.compute_sed_with_extrapolation(times, wavelengths, graph_state, **kwargs)

        # Query the model and convert the output to nJy.
        model_flam = rb_result.get_flux_density(shifted_times, wavelengths)
        model_fnu = flam_to_fnu(
            model_flam,
            wavelengths,
            wave_unit=uu.AA,
            flam_unit=self._FLAM_UNIT,
            fnu_unit=uu.nJy,
        )
        return model_fnu



import numpy as np
from lightcurvelynx.astro_utils.obs_utils import phot_eff_function, spec_eff_function


def test_phot_eff_function():
    """
    test that the phot_eff_function returns correct values.
    """

    snr = [1.0, 3.0, 10.0, 100.0]
    eff = phot_eff_function(snr)
    expected_eff = [0.0, 0.0, 1.0, 1.0]

    np.testing.assert_allclose(eff, expected_eff)


def test_spec_eff_function():
    """
    test the spec efficiency function using data extracted from Figure 4 in Kessler et al. 2019
    """

    imags = [20.01, 21.09, 21.67, 22.54]
    expected_eff = [1.0, 0.9, 0.67, 0.2]
    eff = spec_eff_function(imags)

    np.testing.assert_allclose(eff, expected_eff, atol=0.02)



"""Models that generate a constant SED or bandflux at all times."""

import numpy as np
from astropy import units as u
from citation_compass import cite_function

from lightcurvelynx.math_nodes.given_sampler import GivenValueSampler
from lightcurvelynx.models.physical_model import BandfluxModel, SEDModel
from lightcurvelynx.utils.io_utils import read_numpy_data


class StaticSEDModel(SEDModel):
    """A StaticSEDModel randomly selects an SED at each evaluation and computes
    the flux from that SED at all time steps.

    Parameterized values include:
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - The t0 of the zero phase, date. [from BasePhysicalModel. Not used.]

    Attributes
    ----------
    sed_values : list
       A list of SEDs from which to sample. Each SED is represented as a
       two row numpy-array where the first row is wavelength and the
       second is flux value.

    Parameters
    ----------
    sed_values : list or numpy array.
       A single SED or a list of SEDs from which to sample. Each SED is
       represented as a two row numpy-array where the first row is wavelength
       and the second is flux value.
    weights : numpy.ndarray, optional
        A length N array indicating the relative weight from which to select
        an SED at random. If None, all SEDs will be weighted equally.
    """

    def __init__(
        self,
        sed_values,
        weights=None,
        **kwargs,
    ):
        # If only a single SED was passed, then put it in a list by itself.
        if isinstance(sed_values, np.ndarray) and len(sed_values) == 2:
            self.sed_values = [sed_values]
        else:
            self.sed_values = sed_values

        # Validate the SED input data.
        for idx, sed in enumerate(self.sed_values):
            if not isinstance(sed, np.ndarray) or len(sed.shape) != 2 or sed.shape[0] != 2:
                raise ValueError(f"SED {idx} must be a two row numpy array of wavelength and flux.")
            if sed.shape[1] < 2:
                raise ValueError(f"SED {idx} must have at least 2 entries.")
            if not np.all(np.diff(sed[0, :]) > 0):
                raise ValueError(f"SED {idx} wavelenths are not in sorted order.")

        super().__init__(**kwargs)

        # Create a parameter that indicates which SED was sampled in each simulation.
        all_inds = [i for i in range(len(self.sed_values))]
        self._sampler_node = GivenValueSampler(all_inds, weights=weights)
        self.add_parameter("selected_idx", value=self._sampler_node, allow_gradient=False)

    def __len__(self):
        """Get the number of SED value."""
        return len(self.sed_values)

    @classmethod
    def from_file(cls, sed_file, **kwargs):
        """Load a static SED from a file containing a two column array where the
        first column is wavelength (in angstroms) and the second column is flux (in nJy).

        Parameters
        ----------
        sed_file : str or Path
            The path to the SED file to load.
        **kwargs : dict
            Additional keyword arguments to pass to the StaticSEDModel constructor.

        Returns
        -------
        StaticSEDModel
            An instance of StaticSEDModel with the loaded SED data.
        """
        # Load the SED data from the file (automatically detected format)
        sed_data = read_numpy_data(sed_file)
        if sed_data.ndim != 2 or sed_data.shape[1] != 2:
            raise ValueError(f"SED data from {sed_file} must be a two column array.")

        return cls(sed_values=sed_data.T, **kwargs)

    @classmethod
    @cite_function
    def from_synphot(cls, sp_model, waves=None, **kwargs):
        """Generate the spectrum from a given synphot model.

        References
        ----------
        synphot (ascl:1811.001)

        Parameters
        ----------
        sp_model : synphot.SourceSpectrum
            The synphot model to generate the spectrum from.
        waves : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms) at which to sample the SED.
            If None, the SED will be sampled at the wavelengths defined in the synphot model.
        **kwargs : dict
            Additional keyword arguments to pass to the StaticSEDModel constructor.

        Returns
        -------
        StaticSEDModel
            An instance of StaticSEDModel with the generated SED data.
        """
        try:
            from synphot import units
        except ImportError as err:
            raise ImportError(
                "synphot package is not installed be default. To use the synphot models, please "
                "install it. For example, you can install it with `pip install synphot`."
            ) from err

        if sp_model.z > 0.0:
            raise ValueError(
                "The synphot model must be defined at the rest frame (z=0.0). "
                f"Current redshift is {sp_model.z}."
            )

        if waves is None:
            waves = np.array(sp_model.waveset * u.angstrom)

        # Extract the SED data from the synphot model. Synphot models return flux in units
        # of PHOTLAM (photons s^-1 cm^-2 A^-1), so we convert to nJy.
        photlam_flux = sp_model(waves, flux_unit=units.PHOTLAM)
        sed_data = np.array(units.convert_flux(waves, photlam_flux, "nJy"))
        return cls(np.vstack((waves, sed_data)), **kwargs)

    def minwave(self, graph_state=None):
        """Get the minimum wavelength of the model.

        Parameters
        ----------
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. Not used
            for this model.

        Returns
        -------
        minwave : float or None
            The minimum wavelength of the model (in angstroms) or None
            if the model does not have a defined minimum wavelength.
        """
        idx = self.get_param(graph_state, "selected_idx")
        return self.sed_values[idx][0, 0]

    def maxwave(self, graph_state=None):
        """Get the maximum wavelength of the model.

        Parameters
        ----------
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. Not used
            for this model.

        Returns
        -------
        maxwave : float or None
            The maximum wavelength of the model (in angstroms) or None
            if the model does not have a defined maximum wavelength.
        """
        idx = self.get_param(graph_state, "selected_idx")
        return self.sed_values[idx][0, -1]

    def compute_sed(self, times, wavelengths, graph_state):
        """Draw effect-free observer frame flux densities.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray, optional
            A length N array of observer frame wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of observer frame SED values (in nJy).
        """
        # Use the SED selected by the sampler node to compute the flux density.
        model_ind = self.get_param(graph_state, "selected_idx")
        num_times = len(times)
        num_waves = len(wavelengths)

        # At each time step we interpolate SED at the query wavelengths.
        sed_fluxes = np.interp(
            wavelengths,
            self.sed_values[model_ind][0, :],
            self.sed_values[model_ind][1, :],
            left=0.0,
            right=0.0,
        )

        # We repeat the interpolated SED values at each time.
        flux_density = np.tile(sed_fluxes, num_times).reshape(num_times, num_waves)
        return flux_density


class StaticBandfluxModel(BandfluxModel):
    """A StaticBandfluxModel randomly selects a mapping of bandfluxes at each evaluation
    and uses that at all time steps.

    Parameterized values include:
      * dec - The object's declination in degrees. [from PhysicalModel]
      * distance - The object's luminosity distance in pc. [from PhysicalModel]
      * ra - The object's right ascension in degrees. [from PhysicalModel]
      * redshift - The object's redshift. [from PhysicalModel]
      * t0 - The t0 of the zero phase, date. [from PhysicalModel. Not used.]

    Attributes
    ----------
    bandflux_values : list of dict
       A list of bandflux mappings from which to sample. Each mapping is represented as a
       dictionary where the key is the filter name and the value is the bandflux (in nJy).

    Parameters
    ----------
    bandflux_values : dict or list
       A single bandflux mapping or a list of bandflux mappings from which to sample. Each mapping is
       represented as a dictionary where the key is the filter name and the value is the bandflux (in nJy).
    weights : numpy.ndarray, optional
        A length N array indicating the relative weight from which to select
        a model at random. If None, all models will be weighted equally.
    """

    def __init__(
        self,
        bandflux_values,
        weights=None,
        **kwargs,
    ):
        # If only a single bandflux mapping was passed, then put it in a list by itself.
        if isinstance(bandflux_values, dict):
            self.bandflux_values = [bandflux_values]
        else:
            self.bandflux_values = bandflux_values

        super().__init__(**kwargs)

        # Create a parameter that indicates which bandflux mapping was sampled in each simulation.
        all_inds = [i for i in range(len(self.bandflux_values))]
        self._sampler_node = GivenValueSampler(all_inds, weights=weights)
        self.add_parameter("selected_idx", value=self._sampler_node, allow_gradient=False)

    def __len__(self):
        """Get the number of band flux values."""
        return len(self.bandflux_values)

    def compute_bandflux(self, times, filters, state, rng_info=None):
        """Evaluate the model at the passband level for a single, given graph state.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filters : numpy.ndarray
            A length T array of filter names.
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        """
        # Get the selected bandflux mapping index and values.
        model_ind = self.get_param(state, "selected_idx")
        model_bandflux = self.bandflux_values[model_ind]

        # Fill in the bandflux values corresponding to the filter at each time.
        bandflux = np.zeros(len(times), dtype=float)
        for filter in np.unique(filters):
            filter_mask = filters == filter
            bandflux[filter_mask] = model_bandflux[filter]
        return bandflux



#!/usr/bin/env python

import numpy as np
from astropy.io import fits
from lightcurvelynx import _LIGHTCURVELYNX_TEST_DATA_DIR


def main(args=None):
    """Create fake, small dust maps for testing only."""
    print("Making fake SFD dust map data...")

    # Create a fake map directory
    fake_map_dir = _LIGHTCURVELYNX_TEST_DATA_DIR / "dustmaps" / "sfdmap2"
    if not fake_map_dir.exists():
        fake_map_dir.mkdir(parents=True)

    # Northern Hemisphere Dust
    hdu = fits.PrimaryHDU(np.zeros((2, 2)))
    hdu.header["SIMPLE"] = True
    hdu.header["BITPIX"] = -32
    hdu.header["NAXIS"] = 2
    hdu.header["NAXIS1"] = 2
    hdu.header["NAXIS2"] = 2
    hdu.header["OBJECT"] = "E(B-V)  "
    hdu.header["BUNIT"] = "mag     "
    hdu.header["CRPIX1"] = 0.5
    hdu.header["CRVAL1"] = 90.0
    hdu.header["CTYPE1"] = "GLON-ZEA"
    hdu.header["CRPIX2"] = 0.5
    hdu.header["CRVAL2"] = 90.0
    hdu.header["CTYPE2"] = "GLAT-ZEA"
    hdu.header["CD1_1"] = -0.0395646818624
    hdu.header["CD1_2"] = 0.0
    hdu.header["CD2_1"] = 0.0
    hdu.header["CD2_2"] = 0.0395646818624
    hdu.header["LONPOLE"] = 180
    hdu.header["LAM_NSGP"] = 1
    hdu.header["LAM_SCAL"] = 2
    hdu.writeto(fake_map_dir / "SFD_dust_4096_ngp.fits", overwrite=True)
    print(f"Writing fake SFD dust map data to {fake_map_dir / 'SFD_dust_4096_ngp.fits'}")

    # Southern Hemisphere Dust
    hdu = fits.PrimaryHDU(100.0 * np.ones((2, 2)))
    hdu.header["SIMPLE"] = True
    hdu.header["BITPIX"] = -32
    hdu.header["NAXIS"] = 2
    hdu.header["NAXIS1"] = 2
    hdu.header["NAXIS2"] = 2
    hdu.header["OBJECT"] = "E(B-V)  "
    hdu.header["BUNIT"] = "mag     "
    hdu.header["CRPIX1"] = 0.5
    hdu.header["CRVAL1"] = -90.0
    hdu.header["CTYPE1"] = "GLON-ZEA"
    hdu.header["CRPIX2"] = 0.5
    hdu.header["CRVAL2"] = -90.0
    hdu.header["CTYPE2"] = "GLAT-ZEA"
    hdu.header["CD1_1"] = 0.0395646818624
    hdu.header["CD1_2"] = 0.0
    hdu.header["CD2_1"] = 0.0
    hdu.header["CD2_2"] = -0.0395646818624
    hdu.header["LONPOLE"] = 180
    hdu.header["LAM_NSGP"] = -1
    hdu.header["LAM_SCAL"] = 2
    hdu.writeto(fake_map_dir / "SFD_dust_4096_sgp.fits", overwrite=True)
    print(f"Writing fake SFD dust map data to {fake_map_dir / 'SFD_dust_4096_sgp.fits'}")


if __name__ == "__main__":
    main()



from pathlib import Path

import pytest
from lightcurvelynx import _LIGHTCURVELYNX_TEST_DATA_DIR


@pytest.fixture
def test_data_dir():
    """Return the base test data directory."""
    return Path(_LIGHTCURVELYNX_TEST_DATA_DIR)


@pytest.fixture
def grid_data_good_file(test_data_dir):
    """Return the file path for the good grid input file."""
    return test_data_dir / "grid_input_good.ecsv"


@pytest.fixture
def grid_data_bad_file(test_data_dir):
    """Return the file path for the bad grid input file."""
    return test_data_dir / "grid_input_bad.txt"


@pytest.fixture
def opsim_small(test_data_dir):
    """Return the file path for the small opsim db."""
    return test_data_dir / "opsim_small.db"


@pytest.fixture
def opsim_shorten(test_data_dir):
    """Return the file path for the shortened opsim db."""
    return test_data_dir / "opsim_shorten.db"


@pytest.fixture
def test_flow_filename(test_data_dir):
    """Return the file path for the test flow input file."""
    return test_data_dir / "test_flow.pkl"


@pytest.fixture
def test_conditional_flow_filename(test_data_dir):
    """Return the file path for the test flow input file."""
    return test_data_dir / "test_conditional_flow.pkl"


@pytest.fixture
def oversampled_observations(opsim_shorten):
    """Return an OpSim object with 0.05 day cadence spanning year 2027."""
    from lightcurvelynx.obstable.opsim import OpSim, oversample_opsim

    base_opsim = OpSim.from_db(opsim_shorten)
    return oversample_opsim(
        base_opsim,
        pointing=(0.0, 0.0),
        search_radius=180.0,
        delta_t=0.05,
        time_range=(61406.0, 61771.0),
        bands=None,
        strategy="darkest_sky",
    )


@pytest.fixture
def passbands_dir(test_data_dir):
    """Return the file path for passbands directory."""
    return test_data_dir / "passbands"



"""A wrapper for querying dust maps and returning E(B-V) values.

This module supports a variety of external dustmaps libaries, but
was primarily designed to work with the dustmaps package:
https://github.com/gregreen/dustmaps

Note that the dustmaps package is not included by default, because it uses
dependencies that may not be compatible with all systems. Users can manually
install dustmaps in their environment with “pip install dustmaps” or
“conda install conda-forge::dustmaps”.
"""

import importlib
import logging
from pathlib import Path

import numpy as np
import pooch
from astropy.coordinates import SkyCoord
from citation_compass import CiteClass

from lightcurvelynx import _LIGHTCURVELYNX_BASE_DATA_DIR
from lightcurvelynx.base_models import FunctionNode


class DustEBV(FunctionNode):
    """A wrapper that queries a dust map and returns the ebv for each location.

    This wrapper is designed to work with multiple dust map implementtions
    by providing a stadard interface.

    Attributes
    ----------
    query_fn : function
        The function to query the dust map ebv value given (RA, dec).

    Parameters
    ----------
    query_fn : function
        The function to query the dust map ebv value given (RA, dec).
    RA : parameter
        The object's right ascension (in degrees).
    dec : parameter
        The object's declination (in degrees).
    **kwargs : `dict`, optional
        Any additional keyword arguments.
    """

    def __init__(self, query_fn, ra=None, dec=None, **kwargs):
        # RA and dec are required, but must be passed through as keyword
        # arguments to be used in the sampling function (hence the None default).
        if ra is None or dec is None:
            raise ValueError("RA and dec must be provided to query the dust map.")
        super().__init__(query_fn, ra=ra, dec=dec, outputs=["ebv"], **kwargs)


class DustmapWrapper(DustEBV, CiteClass):
    """A convenience wrapper for the dustmap package
    https://github.com/gregreen/dustmaps

    The DustmapWrapper is designed to take dust map objects from the dustmaps package,
    but can be used with any object that has a query(coords: SkyCoord) method.

    Parameters
    ----------
    dust_map : DustMap object
        The dust map. Since different dustmap's query function may produce different
        outputs, you should include the corresponding ebv_func to transform the result
        into ebv if needed.
    ebv_func : function, optional
        A function to translate the result of the dustmap query into an ebv. This is
        not needed for dustmaps that already produce ebv values.
    RA : parameter
        The object's right ascension (in degrees).
    dec : parameter
        The object's declination (in degrees).
    **kwargs : `dict`, optional
        Any additional keyword arguments.

    References
    ----------
    Green 2018, JOSS, 3(26), 695.
    https://github.com/gregreen/dustmaps
    """

    def __init__(self, dust_map, ra=None, dec=None, ebv_func=None, **kwargs):
        # RA and dec setters are passed through to the DustEBV init method.
        super().__init__(self.compute_ebv, ra=ra, dec=dec, **kwargs)
        self._ebv_func = ebv_func
        self._dust_map = dust_map

    def compute_ebv(self, ra, dec):
        """Compute the E(B-V) value for a given location.

        Parameters
        ----------
        ra : float or np.array
            The object's right ascension (in degrees).
        dec : float or np.array
            The object's declination (in degrees).

        Returns
        -------
        ebv : float or np.array
            The E(B-V) value or array of values.
        """
        if self._dust_map is None:
            raise ValueError("A dust map must be provided to compute E(B-V) values.")

        # Wrap the RA and dec in a SkyCoord object and call the dust map's query method.
        coord = SkyCoord(ra, dec, frame="icrs", unit="deg")
        dustmap_value = self._dust_map.query(coord)

        # Only translate the value to ebv if the dust map does not provide the result in ebv.
        if self._ebv_func is not None:
            return self._ebv_func(dustmap_value)
        return dustmap_value


class ConstantHemisphereDustMap:
    """A DustMap with a constant value in each hemisphere for testing
    and debugging purposes.

    This class is designed to look like a DustMap object from the dustmaps package,
    so that it can be used to test DustmapWrapper.

    Attributes
    ----------
    north_ebv : `float`
        The DustMap's ebv value at all points in the Northern Hemisphere.
    south_ebv : `float`
        The DustMap's ebv value at all points in the Southern Hemisphere.
    """

    def __init__(self, north_ebv, south_ebv):
        self.north_ebv = north_ebv
        self.south_ebv = south_ebv

    def compute_ebv(self, ra, dec):
        """Compute the E(B-V) value for a given location.

        Parameters
        ----------
        ra : float or np.array
            The object's right ascension (in degrees).
        dec : float or np.array
            The object's declination (in degrees).

        Returns
        -------
        ebv : float or np.array
            The E(B-V) value or array of values.
        """
        if np.isscalar(ra):
            return self.north_ebv if dec >= 0 else self.south_ebv

        ebv = np.where(np.asarray(dec) < 0, self.south_ebv, self.north_ebv)
        return ebv

    def query(self, coords):
        """A query function to match the DustMap interface so that we can
        pass ConstantHemisphereDustMap into DustmapWrapper for testing purposes.

        Note
        ----
        This shouldn't be used in production code. Use the compute_ebv()
        function directly.

        Parameters
        ----------
        coords : SkyCoord
            The object's coordinates.

        Returns
        -------
        ebv : float or np.array
            The E(B-V) value or array of values.
        """
        return self.compute_ebv(coords.ra.deg, coords.dec.deg)


class SFDMap(DustEBV):
    """A dustmap using the sfdmap2 package.

    This does not need to be used with DustmapWrapper, but rather is
    a standalone ParameterizedNode.

    Note
    ----
    If the dustmap data is not present in "data/dustmaps/sfdmap2",
    it will automatically be downloaded there when the class is
    instantiated.

    Citations
    ---------
    Software https://github.com/AmpelAstro/sfdmap2
    Forked from https://github.com/kbarbary/sfdmap
    Dust map data from Schlegel, Finkbeiner and Davis (1998).

    Attributes
    ----------
    dustmap : sfdmap.SFDMap
        The dust map object.

    Parameters
    ----------
    data_dir : `str`, optional
        The directory containing the dust map data files.
        If None, the default directory will be used.
    RA : parameter
        The object's right ascension (in degrees).
    dec : parameter
        The object's declination (in degrees).
    **kwargs : `dict`, optional
        Any additional keyword arguments.
    """

    _default_map_dir = _LIGHTCURVELYNX_BASE_DATA_DIR / "dustmaps" / "sfdmap2"

    def __init__(self, data_dir=None, ra=None, dec=None, **kwargs):
        # Pass RA and dec through to DustEBV's init method.
        super().__init__(self.compute_ebv, ra=ra, dec=dec, **kwargs)
        self.dustmap = self._load_data(data_dir)

    def _data_files_exist(self, data_dir):
        """Check that the necessary data files exist in the given directory."""
        if not data_dir.exists() or not data_dir.is_dir():
            return False
        elif (data_dir / "SFD_dust_4096_ngp.fits").exists() is False:
            return False
        elif (data_dir / "SFD_dust_4096_sgp.fits").exists() is False:
            return False
        return True

    def _load_data(self, data_dir=None):
        """Load the dust map data files.

        Parameters
        ----------
        data_dir : `str`, optional
            The directory containing the dust map data files.
            If None, the default directory will be used.

        Returns
        -------
        dustmap : sfdmap.SFDMap
            The dust map object.
        """
        logger = logging.getLogger(__name__)

        # Check that we have the sfdmap2 package installed (since it is not installed by default).
        if importlib.util.find_spec("sfdmap2") is None:
            raise ImportError(
                "The sfdmap2 package is required to use the SFDMap effect and not installed by default. "
                "You can install sfdmap2 using `pip install sfdmap2`."
            )
        else:
            from sfdmap2 import sfdmap

        data_dir = Path(data_dir) if data_dir is not None else self._default_map_dir
        logger.debug(f"Loading SFD dust map data from {data_dir}")

        if not self._data_files_exist(data_dir):
            data_url = ("https://github.com/kbarbary/sfddata/archive/master.tar.gz",)
            logger.info(
                "SFD dust map data files not found.\n"
                f"Attempting to download from: {data_url}\n"
                f"to the directory {data_dir}"
            )

            # Create the data directory if it doesn't exist.
            if not data_dir.exists():
                data_dir.mkdir(parents=True)

            # Use pooch to download the data files and extract them to the data directory.
            pooch.retrieve(
                url="https://github.com/kbarbary/sfddata/archive/master.tar.gz",
                known_hash="95e8645dcdcbd4ad48398d44307741f550bdde95e8f438b2a3be0021723a4d7e",
                processor=pooch.Untar(extract_dir=data_dir),
            )
            data_dir = data_dir / "sfddata-master"

        # Check that the files were downloaded and extracted.
        if not self._data_files_exist(data_dir):
            raise ValueError(f"The SFD dust map data files are missing from {data_dir}.")

        return sfdmap.SFDMap(data_dir)

    def compute_ebv(self, ra, dec):
        """Compute the E(B-V) value for a given location.

        Parameters
        ----------
        ra : float or np.array
            The object's right ascension (in degrees).
        dec : float or np.array
            The object's declination (in degrees).

        Returns
        -------
        ebv : float or np.array
            The E(B-V) value or array of values.
        """
        ebv = self.dustmap.ebv(ra, dec)
        return ebv



import numpy as np
from lightcurvelynx.effects.microlensing import Microlensing
from lightcurvelynx.models.basic_models import ConstantSEDModel


def test_microlensing() -> None:
    """Test that we can apply a basic microlensing effect."""
    num_times = 50
    num_waves = 20
    times = np.arange(num_times, dtype=float)

    microlensing_t0 = 5.0
    microlensing = Microlensing(microlensing_t0=microlensing_t0, u_0=0.1, t_E=10.0)

    # Apply at the wavelength level. Note that when calling apply() manually, we
    # need to pass along the parameter values.
    wave_fluxes = np.full((num_times, num_waves), 100.0)
    values = microlensing.apply(
        wave_fluxes,
        times=times,
        apply_microlensing=True,
        microlensing_t0=microlensing_t0,
        u_0=0.1,
        t_E=10.0,
    )

    # Check that the fluxes are all magnified (>100) and that they increase
    # before t0 and decrease after it.
    assert values.shape == (num_times, num_waves)
    assert np.all(values >= 100.0)
    for wave_idx in range(num_waves):
        assert np.all(np.diff(values[0:5, wave_idx]) >= 0)
        assert np.all(np.diff(values[5:num_times, wave_idx]) <= 0)


def test_microlensing_bandflux() -> None:
    """Test that we can apply a basic microlensing effect at the bandflux level."""
    num_times = 50
    times = np.arange(num_times, dtype=float)

    microlensing_t0 = 5.0
    microlensing = Microlensing(microlensing_t0=microlensing_t0, u_0=0.1, t_E=10.0)

    # Apply at the wavelength level. Note that when calling apply() manually, we
    # need to pass along the parameter values.
    bandfluxes = np.full(num_times, 100.0)
    values = microlensing.apply_bandflux(
        bandfluxes,
        times=times,
        apply_microlensing=True,
        microlensing_t0=microlensing_t0,
        u_0=0.1,
        t_E=10.0,
    )

    # Check that the fluxes are all magnified (>100) and that they increase
    # before t0 and decrease after it.
    assert values.shape == (num_times,)
    assert np.all(values >= 100.0)
    assert np.all(np.diff(values[0:5]) >= 0)
    assert np.all(np.diff(values[5:num_times]) <= 0)


def test_constant_sed_microlensing() -> None:
    """Test that we can apply microlensing to a ConstantSEDModel."""
    num_times = 50
    microlensing_t0 = 5.0
    times = np.arange(num_times, dtype=float)

    model = ConstantSEDModel(brightness=10.0, node_label="my_constant_sed_model")
    microlensing = Microlensing(microlensing_t0=microlensing_t0, u_0=0.1, t_E=10.0)
    model.add_effect(microlensing)

    state = model.sample_parameters()
    wavelengths = np.array([100.0, 200.0, 300.0])
    values = model.evaluate_sed(times, wavelengths, state)
    assert values.shape == (num_times, 3)

    # Check that the fluxes are all magnified (>100) and that they increase
    # before t0 and decrease after it.
    assert values.shape == (num_times, 3)
    assert np.all(values >= 10.0)
    for wave_idx in range(3):
        assert np.all(np.diff(values[0:5, wave_idx]) >= 0)
        assert np.all(np.diff(values[5:num_times, wave_idx]) <= 0)



"""Utility functions for post processing the results data by adding statistics
columns and filtering on those columns."""

import numpy as np
import numpy.ma as ma
import pandas as pd
from nested_pandas import NestedFrame

from lightcurvelynx.astro_utils.mag_flux import flux2mag


def concat_results(results_list):
    """Concatenate a list of results into a single NestedFrame,
    updating the ID column to be unique across all results.

    Parameters
    ----------
    results_list : list of nested_pandas.NestedFrame
        The list of DataFrames to concatenate.

    Returns
    -------
    nested_pandas.NestedFrame
        The concatenated DataFrame.
    """
    result = pd.concat(results_list, ignore_index=True)

    # We need to update the ID column to be unique across all results.
    if "id" in result.columns:
        result["id"] = np.arange(len(result))
    return result


def results_drop_empty(results):
    """Drop empty lightcurves from the results DataFrame.

    Parameters
    ----------
    results : nested_pandas.NestedFrame
        The DataFrame containing lightcurve data.

    Returns
    -------
    nested_pandas.NestedFrame
        The DataFrame with empty lightcurves removed.
    """
    return results.dropna(subset=["lightcurve"])


def lightcurve_compute_snr(flux, fluxerr):
    """Compute the signal-to-noise ratio (SNR) for given flux and flux error arrays.

    Parameters
    ----------
    flux : array-like
        The flux values.
    fluxerr : array-like
        The flux error values.

    Returns
    -------
    result : np.ndarray
        The SNR values, with None for invalid entries (e.g., zero or negative flux error).
    """
    flux = np.asarray(flux)
    fluxerr = np.asarray(fluxerr)
    valid_mask = (flux > 0) & (fluxerr > 0)

    result = ma.masked_all(flux.shape)
    result[valid_mask] = flux[valid_mask] / fluxerr[valid_mask]
    return result


def lightcurve_compute_mag(flux, fluxerr):
    """Compute the AB magnitude and magnitude error for given flux and flux error arrays.

    Parameters
    ----------
    flux : array-like
        The flux values.
    fluxerr : array-like
        The flux error values.

    Returns
    -------
    tuple of np.ndarray
        The magnitude and magnitude error values, with None for invalid entries (e.g., non-positive flux).
    """
    flux = np.asarray(flux)
    fluxerr = np.asarray(fluxerr)
    valid_mask = (flux > 0) & (fluxerr > 0)

    mag = ma.masked_all(flux.shape)
    mag[valid_mask] = flux2mag(flux[valid_mask])

    magerr = ma.masked_all(flux.shape)
    magerr[valid_mask] = (2.5 / np.log(10)) * (fluxerr[valid_mask] / flux[valid_mask])

    return mag, magerr


def augment_single_lightcurve(results, *, min_snr=0.0, t0=None):
    """Add columns to a single lightcurve DataFrame with additional information
    about the light curve, including:
    - SNR = flux / fluxerr
    - detection flag (True if SNR >= min_snr, False otherwise)
    - AB magnitude
    - AB magnitude error = (2.5 / ln(10)) * (fluxerr / flux)
    - relative time = mjd - t0 (if t0 is provided)
    None is used for invalid entries, e.g. negative flux or zero flux error.

    Parameters
    ----------
    results : pandas.DataFrame
        The DataFrame containing lightcurve data. Modified in place.
    min_snr : float, optional
        Minimum SNR required to mark an entry as a detection. Default is 0.0.
    t0 : float or None, optional
        Reference time for the lightcurve.
    """
    if "flux" not in results.columns or "fluxerr" not in results.columns:
        raise ValueError("flux and fluxerr must be present in the light curve DataFrame.")
    flux = results["flux"]
    fluxerr = results["fluxerr"]

    snr = lightcurve_compute_snr(flux, fluxerr)
    results["snr"] = snr
    results["detection"] = [(x is not ma.masked and x >= min_snr) for x in snr]

    mag, magerr = lightcurve_compute_mag(flux, fluxerr)
    results["mag"] = mag
    results["magerr"] = magerr

    if t0 is not None and "mjd" in results.columns:
        results["time_rel"] = results["mjd"] - t0


def results_augment_lightcurves(results, *, min_snr=0.0):
    """Add columns to the results DataFrame with additional information
    about each light curve, including:
    - SNR = flux / fluxerr
    - detection flag (True if SNR >= min_snr, False otherwise)
    - AB magnitude
    - AB magnitude error = (2.5 / ln(10)) * (fluxerr / flux)
    - relative time = mjd - t0 (if t0 in the results table)
    None is used for invalid entries, e.g. negative flux or zero flux error.

    The input data frame can either be a single light curve (pandas.DataFrame)
    with columns "flux" and "fluxerr", or a NestedFrame (nested_pandas.NestedFrame)
    with a nested DataFrame column "lightcurve" that contains the "flux" and
    "fluxerr" columns.

    Parameters
    ----------
    results : pandas.DataFrame or nested_pandas.NestedFrame
        The DataFrame containing lightcurve data. Modified in place.
    min_snr : float, optional
        Minimum SNR required to mark an entry as a detection. Default is 0.0.
    """
    if not isinstance(results, NestedFrame) or "lightcurve" not in results.columns:
        raise ValueError("results must be a NestedFrame with a 'lightcurve' column.")
    if "flux" not in results["lightcurve"].nest.fields or "fluxerr" not in results["lightcurve"].nest.fields:
        raise ValueError("lightcurve.flux and lightcurve.fluxerr must be present in the DataFrame.")
    flux = results["lightcurve.flux"]
    fluxerr = results["lightcurve.fluxerr"]

    # Compute SNR and detection flag.
    snr = lightcurve_compute_snr(flux, fluxerr)
    results["lightcurve.snr"] = snr
    results["lightcurve.detection"] = [(x is not None and x >= min_snr) for x in snr]

    # Compute magnitude and magnitude error.
    mag, magerr = lightcurve_compute_mag(flux, fluxerr)
    results["lightcurve.mag"] = mag
    results["lightcurve.magerr"] = magerr

    # If t0 is provided as a column in results, compute relative time.
    if "t0" in results.columns and np.all(results["t0"]) and results["t0"].notna().all():
        if "mjd" not in results["lightcurve"].nest.fields:
            raise ValueError("lightcurve.mjd must be present in the DataFrame.")

        # Get the index for the t0 entry for each lightcurve MJD and use that
        # to subtract out the reference t0.
        t0 = np.asanyarray(results["t0"])
        t0_idx = np.array(results["lightcurve"]["mjd"].index)
        results["lightcurve.time_rel"] = results["lightcurve.mjd"] - t0[t0_idx]



import sqlite3
from pathlib import Path

import numpy as np
import pandas as pd
from astropy.time import Time

from lightcurvelynx.astro_utils.mag_flux import mag2flux
from lightcurvelynx.astro_utils.noise_model import poisson_bandflux_std
from lightcurvelynx.astro_utils.zeropoint import calculate_zp_from_maglim
from lightcurvelynx.consts import GAUSS_EFF_AREA2FWHM_SQ
from lightcurvelynx.obstable.obs_table import ObsTable

ZTFCAM_PIXEL_SCALE = 1.01
"""The pixel scale for the ZTF camera in arcseconds per pixel."""

_ztfcam_readout_noise = 8
"""The standard deviation of the count of readout electrons per pixel for the ZTF camera."""

_ztfcam_dark_current = 0.0
"""The dark current for the ZTF camera in electrons per second per pixel."""

_ztfcam_view_radius = 2.735
"""The angular radius of the observation field (in degrees). ZTF FOV is 47 deg^2"""

_ztfcam_ccd_gain = 6.2
"""CCD gain (in e-/ADU)"""


class ZTFObsTable(ObsTable):
    """A subclass for ZTF exposure table.

    Parameters
    ----------
    table : dict or pandas.core.frame.DataFrame
        The table with all the observation information.
    colmap : dict
        A mapping of short column names to their names in the underlying table.
        Defaults to the ZTF column names, stored in _default_colnames.
    **kwargs : dict
        Additional keyword arguments to pass to the ObsTable constructor. This includes overrides
        for survey parameters such as:
        - dark_current : The dark current for the camera in electrons per second per pixel.
        - gain: The CCD gain (in e-/ADU).
        - pixel_scale: The pixel scale for the camera in arcseconds per pixel.
        - radius: The angular radius of the observations (in degrees).
        - read_noise: The standard deviation of the count of readout electrons per pixel.
    """

    # Default column names for the ZTF survey data.
    _default_colnames = {
        "maglim": "maglim",
        "sky": "scibckgnd",
        "fwhm": "fwhm",
        "dec": "dec",
        "exptime": "exptime",
        "filter": "filter",
        "ra": "ra",
        "time": "obsmjd",
        "zp": "zp_nJy",  # We add this column to the table
    }

    # Default survey values.
    _default_survey_values = {
        "dark_current": _ztfcam_dark_current,
        "gain": _ztfcam_ccd_gain,
        "pixel_scale": ZTFCAM_PIXEL_SCALE,
        "radius": _ztfcam_view_radius,
        "read_noise": _ztfcam_readout_noise,
        "survey_name": "ZTF",
    }

    def __init__(self, table, colmap=None, **kwargs):
        colmap = self._default_colnames if colmap is None else colmap

        # Make a copy of the table data with the obsdate converted to the MJD and
        # save in time.
        if "obsdate" in table and "obsmjd" not in table:
            table = table.copy()
            t = Time(list(table["obsdate"]), format="iso", scale="utc")
            table["obsmjd"] = t.mjd

        super().__init__(table, colmap=colmap, **kwargs)

    def _assign_zero_points(self):
        """Assign instrumental zero points in ADU to the ObsTable."""
        cols = self._table.columns.tolist()
        if not ("maglim" in cols and "sky" in cols and "fwhm" in cols and "exptime" in cols):
            raise ValueError(
                "ObsTable does not include the columns needed to derive zero point "
                "information. Required columns: maglim, sky, fwhm and exptime."
            )

        # replace invalid values in table
        self._table = self._table.replace("", np.nan)
        self._table = self._table.dropna(subset=["fwhm"])

        zp_values = calculate_zp_from_maglim(
            maglim=self._table["maglim"],
            sky=self._table["sky"],
            fwhm=self._table["fwhm"],
            gain=_ztfcam_ccd_gain,
            readnoise=_ztfcam_readout_noise,
            darkcurrent=_ztfcam_dark_current,
            exptime=self._table["exptime"],
            nexposure=1,
        )
        zp_nJy = mag2flux(zp_values)
        self.add_column("zp", zp_nJy, overwrite=True)

    @classmethod
    def from_db(cls, filename, sql_query="SELECT * from exposures", colmap=None):
        """Create an ObsTable object from the data in a db file.

        Parameters
        ----------
        filename : str
            The name of the db file.
        sql_query : str
            The SQL query to use when loading the table.
            Default: "SELECT * FROM observations"
        colmap : dict, optional
            A mapping of short column names to their names in the underlying table.
            If None then defaults to the ZTF column names.

        Returns
        -------
        obstable : ZTFObsTable
            A table with all of the pointing data.

        Raise
        -----
        FileNotFoundError if the file does not exist.
        ValueError if unable to load the table.
        """
        if colmap is None:
            colmap = cls._default_colnames

        if not Path(filename).is_file():
            raise FileNotFoundError(f"ObsTable file {filename} not found.")
        con = sqlite3.connect(f"file:{filename}?mode=ro", uri=True)

        # Read the table.
        try:
            obstable = pd.read_sql_query(sql_query, con)
        except Exception:
            raise ValueError("ObsTable database read failed.") from None

        # Close the connection.
        con.close()

        return ZTFObsTable(obstable, colmap=colmap)

    def bandflux_error_point_source(self, bandflux, index):
        """Compute observational bandflux error for a point source

        Parameters
        ----------
        bandflux : array_like of float
            Band bandflux of the point source in nJy.
        index : array_like of int
            The index of the observation in the table.

        Returns
        -------
        flux_err : array_like of float
            Simulated bandflux noise in nJy.
        """
        observations = self._table.iloc[index]

        # By the effective FWHM definition, see
        # https://smtn-002.lsst.io/v/OPSIM-1171/index.html
        footprint = GAUSS_EFF_AREA2FWHM_SQ * observations["fwhm"] ** 2  # in pixels

        return poisson_bandflux_std(
            bandflux,  # nJy
            total_exposure_time=observations["exptime"],
            exposure_count=1,
            psf_footprint=footprint,
            sky=observations["sky"] * self.safe_get_survey_value("gain"),  # e-/pixel^2
            zp=observations["zp"],  # nJy
            readout_noise=self.safe_get_survey_value("read_noise"),  # e-/pixel
            dark_current=self.safe_get_survey_value("dark_current"),  # e-/second/pixel
        )


def create_random_ztf_obs_data(num_obs, seed=None):
    """Create a random ObsTable pointings drawn uniformly from (RA, dec).

    Parameters
    ----------
    num_obs : int
        The size of the ObsTable to generate.
    seed : int
        The seed to used for random number generation. If None then
        uses a default random number generator.
        Default: None

    Returns
    -------
    obstable : pd.DataFrame
        The data for the ObsTable.
    """
    if num_obs <= 0:
        raise ValueError("Number of observations must be greater than zero.")

    rng = np.random.default_rng() if seed is None else np.random.default_rng(seed=seed)

    # Generate the (RA, dec) pairs uniformly on the surface of a sphere.
    ra = np.degrees(rng.uniform(0.0, 2.0 * np.pi, size=num_obs))
    dec = np.degrees(np.arccos(2.0 * rng.uniform(0.0, 1.0, size=num_obs) - 1.0) - (np.pi / 2.0))

    # Generate the information needed to compute zeropoint.
    maglim = rng.uniform(19.0, 21.0, size=num_obs)
    sky = rng.uniform(100.0, 200.0, size=num_obs)
    fwhm = rng.uniform(1.0, 3.0, size=num_obs)
    filter = rng.choice(["g", "r", "i"], size=num_obs)

    input_data = {
        "obsdate": ["2018-03-25 06:04:40.000"] * num_obs,
        "ra": ra,
        "dec": dec,
        "maglim": maglim,
        "scibckgnd": sky,
        "fwhm": fwhm,
        "filter": filter,
        "exptime": 30.0 * np.ones(num_obs),
    }
    return pd.DataFrame(input_data)



import numpy as np
from astropy.cosmology import WMAP9, Planck18
from lightcurvelynx.astro_utils.redshift import RedshiftDistFunc, redshift_to_distance
from lightcurvelynx.models.basic_models import StepModel


def test_redshifted_flux_densities() -> None:
    """Test that we correctly calculate redshifted values."""
    times = np.linspace(0, 100, 1000)
    wavelengths = np.array([100.0, 200.0, 300.0])
    t0 = 10.0
    t1 = 30.0
    brightness = 50.0

    for redshift in [0.0, 0.5, 2.0, 3.0, 30.0]:
        model_redshift = StepModel(brightness=brightness, t0=t0, t1=t1, redshift=redshift)
        values_redshift = model_redshift.evaluate_sed(times, wavelengths)

        for i, time in enumerate(times):
            if t0 <= time and time <= (t1 - t0) * (1 + redshift) + t0:
                assert np.all(values_redshift[i] == brightness * (1 + redshift))
            else:
                assert np.all(values_redshift[i] == 0.0)


def test_redshift_to_distance():
    """Test that we can convert the redshift to a distance using a given cosmology."""
    wmap9_val = redshift_to_distance(1100, cosmology=WMAP9)
    planck18_val = redshift_to_distance(1100, cosmology=Planck18)

    assert abs(planck18_val - wmap9_val) > 1000.0
    assert 13.0 * 1e12 < wmap9_val < 16.0 * 1e12
    assert 13.0 * 1e12 < planck18_val < 16.0 * 1e12


def test_redshift_dist_func_node():
    """Test the RedshiftDistFunc node."""
    node = RedshiftDistFunc(redshift=1100, cosmology=Planck18)
    state = node.sample_parameters()
    assert 13.0 * 1e12 < node.get_param(state, "function_node_result") < 16.0 * 1e12

    # Test that we can generate multiple samples.
    state = node.sample_parameters(num_samples=10)
    assert np.all(node.get_param(state, "function_node_result") > 13.0 * 1e12)
    assert np.all(node.get_param(state, "function_node_result") < 16.0 * 1e12)



import astropy.units as u
import numpy as np
import scipy.integrate as integrate
from astropy.coordinates import SkyCoord
from astropy.cosmology import FlatLambdaCDM
from scipy.stats import norm
from scipy.stats.sampling import NumericalInversePolynomial

from lightcurvelynx.base_models import FunctionNode
from lightcurvelynx.math_nodes.scipy_random import NumericalInversePolynomialFunc


def snia_volumetric_rates(redshift):
    """
    SN Ia volumetric rate based on Frohmaier et al. (2019).
    r_v(z) = r0 * (1+z)^alpha （SNe Ia yr^-1 Mpc^-3 h_70^3）
    r0 = 2.27+/-0.19e-5
    alpha = 1.7+/-0.21

    Parameters
    ----------
    redshift: float or numpy.ndarray
        The redshift of the supernova

    Returns
    -------
    rate_vol: float or numpy.ndarray
        The volumetric rate of the supernova given the redshift
    """

    r0 = 2.27e-5
    alpha = 1.7
    rate_vol = r0 * np.power(1.0 + redshift, alpha)

    return rate_vol


def num_snia_per_redshift_bin(zmin=0.001, zmax=10, znbins=20, solid_angle=None, H0=73.0, Omega_m=0.3):
    """
    Calculate the number of SNe Ia in each redshift bin based on rates.

    r_v(z) = dN/dz
    V = comoving volume
    T = length of survey in years
    N = int r_v(z)dz * dV * dT

    Parameters
    ----------
    zmin: float
        Min redshift value for calculation.
    zmax: float
        Max redshift value for calculation.
    znbins: int
        Number of redshift bins for calculating SNe Ia numbers.
    solid_angle: float
        Solid angle for calculating the number of SNe (in sr).
    H0: float
        The Hubble Constant.
    Omega_m: float
        The matter density.

    Returns
    -------
    num_sn: numpy.ndarray
        Number of SNe Ia in each zbin per year.
    z_mean: numpy.ndarray
        Mean value for each redshift bin.
    """

    if solid_angle is None:
        solid_angle = 4 * np.pi

    cosmo = FlatLambdaCDM(H0=H0, Om0=Omega_m)

    zarr = np.linspace(zmin, zmax, znbins + 1)

    int_arr = np.linspace(zarr[:-1], zarr[1:], 50, axis=1)
    z_mean = np.mean(int_arr, axis=1)

    dV = (
        solid_angle * cosmo.differential_comoving_volume(int_arr) * (H0 / 70.0) ** 3
    )  # * 4pi because differential_comoving_volume is per solid angle
    r_v = snia_volumetric_rates(int_arr)
    dn_dz = r_v * dV.value

    num_sn = integrate.trapezoid(dn_dz, int_arr, axis=1)

    return num_sn, z_mean


class HostmassX1Distr:
    """
    A class that contains the pdf of the SALT x1 parameter given the hostmass

    Attributes
    ----------
    hostmass: float
        The hostmass value.

    Parameters
    ----------
    hostmass: float
        The hostmass value.
    """

    def __init__(self, hostmass):
        self.hostmass = hostmass

    def _p(self, x1, hostmass=9.0):
        """
        The probablity of having a value of x1 given a hostmass.

        Parameters
        ----------
        x1: numpy.ndarray
            The x1 value.
        hostmass: float
            The hostmass value.

        Returns
        -------
        p: numpy.ndarray
            The probablity.
        """

        p = np.exp(-(np.minimum(0, x1) ** 2))
        p = np.where(np.logical_and(x1 > -5, x1 < 5), p, 0.0)
        p = np.where(hostmass < 10.0, p, 1.0)

        return p

    def pdf(self, x1):
        """
        The pdf of x1 given hostmass.

        Parameters
        ----------
        x1: numpy.ndarray
            The x1 value.
        hostmass: float
            The hostmass value.

        Returns
        -------
        The pdf function of x1 given hostmass.

        """
        return self._p(x1, hostmass=self.hostmass) * norm.pdf(x1, loc=0, scale=1)


def _x0_from_distmod(distmod, x1, c, alpha, beta, m_abs):
    """Calculate the SALT3 x0 parameter given distance modulus based on Tripp relation.
    distmod = -2.5*log10(x0) + alpha * x1 - beta * c - m_abs + 10.635
    x0 = 10 ^ (-0.4* (distmod - alpha * x1 + beta * c + m_abs - 10.635))

    Parameters
    ----------
    distmod : float
        The distance modulus value (in mag).
    x1 : float
        The SALT3 x1 parameter.
    c : float
        The SALT3 c parameter.
    alpha : float
        The alpha parameter in the Tripp relation.
    beta : float
        The beta parameter in the Tripp relation.
    m_abs : float
        The absolute magnitude of SN Ia.

    Returns
    -------
    x0 : float
        The x0 parameter
    """
    x0 = np.power(10.0, -0.4 * (distmod - alpha * x1 + beta * c + m_abs - 10.635))

    return x0


class HostmassX1Func(NumericalInversePolynomialFunc):
    """A class for sampling from the HostmassX1Distr.

    Parameters
    ----------
    hostmass : function or constant
        The function or constant providing the hostmass value.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, hostmass, **kwargs):
        # Since HostmassX1Distr can only take on two values (one for hostmass < 10.0 and one
        # for hostmass >= 10.0), we just create two distributions.
        self._dist_ge_10 = HostmassX1Distr(11.0)
        self._dist_lt_10 = HostmassX1Distr(9.0)
        super().__init__(
            dist=HostmassX1Distr,
            hostmass=hostmass,
            **kwargs,
        )

        # We override the inverse functions.
        self._inv_poly_ge_10 = NumericalInversePolynomial(self._dist_ge_10)
        self._inv_poly_lt_10 = NumericalInversePolynomial(self._dist_lt_10)
        self._vect_sample = None

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Sample from one of the two distributions depending on hostmass.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator or None, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.
        """
        rng = rng_info if rng_info is not None else self._rng
        hostmass = self.get_param(graph_state, "hostmass")

        if graph_state.num_samples == 1:
            results = (
                self._inv_poly_ge_10.rvs(1, rng)[0] if hostmass >= 10 else self._inv_poly_lt_10.rvs(1, rng)[0]
            )
        else:
            results = np.zeros(graph_state.num_samples)

            # Batch generate samples for all points with hostmass >= 10.0
            ge_10_idx = hostmass >= 10.0
            results[ge_10_idx] = self._inv_poly_ge_10.rvs(np.count_nonzero(ge_10_idx), rng)

            # Batch generate samples for all points with hostmass < 10.0
            lt_10_idx = hostmass < 10.0
            results[lt_10_idx] = self._inv_poly_lt_10.rvs(np.count_nonzero(lt_10_idx), rng)

        self._save_results(results, graph_state)
        return results


class X0FromDistMod(FunctionNode):
    """A wrapper class for the _x0_from_distmod() function.

    Parameters
    ----------
    distmod : function or constant
        The function or constant providing the distance modulus value.
    x1 : function or constant
        The function or constant providing the x1 value.
    c : function or constant
        The function or constant providing the c value.
    alpha : function or constant
        The function or constant providing the alpha value.
    beta : function or constant
        The function or constant providing the beta value.
    m_abs : function or constant
        The function or constant providing the m_abs value.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, distmod, x1, c, alpha, beta, m_abs, **kwargs):
        # Call the super class's constructor with the needed information.
        super().__init__(
            func=_x0_from_distmod,
            distmod=distmod,
            x1=x1,
            c=c,
            alpha=alpha,
            beta=beta,
            m_abs=m_abs,
            **kwargs,
        )


class DistModFromRedshift(FunctionNode):
    """A wrapper class for the _distmod_from_redshift() function.

    Parameters
    ----------
    redshift : function or constant
        The function or constant providing the redshift value.
    H0 : constant
        The Hubble constant.
    Omega_m : constant
        The matter density Omega_m.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, redshift, H0=73.0, Omega_m=0.3, **kwargs):
        # Create the cosmology once for this node.
        if not isinstance(H0, float) or not isinstance(Omega_m, float):
            raise ValueError("H0 and Omega_m must be constants.")
        self.cosmo = FlatLambdaCDM(H0=H0, Om0=Omega_m)

        # Call the super class's constructor with the needed information.
        super().__init__(
            func=self._distmod_from_redshift,
            redshift=redshift,
            **kwargs,
        )

    def _distmod_from_redshift(self, redshift):
        """Compute distance modulus given redshift and cosmology.

        Parameters
        ----------
        redshift : float or numpy.ndarray
            The redshift value(s).

        Returns
        -------
        distmod : float or numpy.ndarray
            The distance modulus (in mag)
        """
        return self.cosmo.distmod(redshift).value


class SNCoordGivenPhysicalSep(FunctionNode):
    """A class for gernerating SN Coordinates given host coordinates and physical separations.

    Parameters
    ----------
    host_ra : function or constant
        Host galaxy RA in degree.
    host_dec : function or constant
        Host galaxy DEC in degree.
    physical_sep_kpc : function or constant
        The physical separation between host and SN in kpc.
    redshift : function or constant
        The function or constant providing the redshift value.
    H0 : constant
        The Hubble constant.
    Omega_m : constant
        The matter density Omega_m.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, host_ra, host_dec, physical_sep_kpc, redshift, H0=73.0, Omega_m=0.3, **kwargs):
        # Create the cosmology once for this node.
        if not isinstance(H0, float) or not isinstance(Omega_m, float):
            raise ValueError("H0 and Omega_m must be constants.")
        self.cosmo = FlatLambdaCDM(H0=H0, Om0=Omega_m)

        # Call the super class's constructor with the needed information.
        super().__init__(
            func=self._sn_coord,
            host_ra=host_ra,
            host_dec=host_dec,
            physical_sep_kpc=physical_sep_kpc,
            redshift=redshift,
            outputs=["ra", "dec"],
            **kwargs,
        )

    def _host_sn_angular_separation(self, physical_sep_kpc, redshift):
        """
        Function for calculating host-sn angular separation.

        Parameters
        ----------
        physical_sep_kpc : float or numpy.ndarray
            The physical host-sn separation(s) in kpc.
        redshift: float or numpy.ndarray
            The redshift value(s).

        Returns
        -------
        angular_sep_arcsec : float or numpy.ndarray
            The angular host-sn separation(s) in arc seconds.
        """

        angular_diameter_distance = self.cosmo.angular_diameter_distance(redshift)
        angular_sep_rad = (physical_sep_kpc * u.kpc / angular_diameter_distance) * u.rad

        return angular_sep_rad

    def _sn_coord(self, host_ra, host_dec, physical_sep_kpc, redshift):
        """
        Function to generate SN coordinates given the host coordinates and angular separation.

        Parameters
        ----------
        host_ra : float or numpy.ndarray
            Host galaxy RA values (in degrees).
        host_dec : float or numpy.ndarray
            Host galaxy DEC values (in degrees).
        physical_sep_kpc : float or numpy.ndarray
            The physical host-sn separation(s) in kpc.
        redshift : float or numpy.ndarray
            The redshift to convert physical separation to angular separation.

        Returns
        -------
        ra : float or numpy.ndarray
            SN RA values (in degrees).
        dec : float or numpy.ndarray
            SN DEC values (in degrees).
        """

        self.angular_sep_rad = self._host_sn_angular_separation(physical_sep_kpc, redshift)
        center = SkyCoord(host_ra * u.deg, host_dec * u.deg, frame="icrs")
        rand_pa = 2 * np.pi * np.random.random(center.size) * u.rad  # random position angle

        sn_coord = center.directional_offset_by(rand_pa, self.angular_sep_rad)
        ra = sn_coord.ra.deg
        dec = sn_coord.dec.deg

        return ra, dec



"""Wrapper classes for some of scipy's sampling functions."""

from os import urandom

import numpy as np
from scipy.stats.sampling import NumericalInversePolynomial

from lightcurvelynx.base_models import FunctionNode
from lightcurvelynx.graph_state import transpose_dict_of_list


class NumericalInversePolynomialFunc(FunctionNode):
    """A class for sampling from scipy's NumericalInversePolynomial
    given a distribution function, an object with a pdf function,
    or a class from which to create such an object.

    Note
    ----
    If a class is provided, then the sampling function will create a new
    object (with the sampled parameters) for each sampling. This is very expensive.

    Attributes
    ----------
    _dist : object or class
        An object or class with either a pdf() or logpdf() method that defines
        the distribution from which to sample.
    _inv_poly: scipy.stats.sampling.NumericalInversePolynomial
        The scipy object to use for sampling. Set to None if _dist is a class.
    _vect_sample : numpy.vectorize
        The vectorized function to create a distribution from a class and sample it.
        Set to None if _dist is an object.
    _rng : numpy.random._generator.Generator
        This object's random number generator.

    Parameters
    ----------
    dist : object or class
        An object or class with either a pdf() or logpdf() method that defines
        the distribution from which to sample.
    seed : int, optional
        The seed to use.
    """

    def __init__(self, dist=None, seed=None, **kwargs):
        # Check that the distribution object/class has a pdf or logpdf function
        # or that we have provided a function directly.
        if not hasattr(dist, "pdf") and not hasattr(dist, "logpdf"):
            raise ValueError("Distribution must have either pdf() or logpdf().")
        self._dist = dist

        # Classes show up as type="type". In this case we will need to create
        # a concrete object from the class and any given parameters.
        if isinstance(dist, type):
            self._inv_poly = None
            self._vect_sample = np.vectorize(self._create_and_sample)
        else:
            self._inv_poly = NumericalInversePolynomial(self._dist)
            self._vect_sample = None

        # Get a default random number generator for this object, using the
        # given seed if one is provided.
        if seed is None:
            seed = int.from_bytes(urandom(4), "big")
        self._rng = np.random.default_rng(seed=seed)

        # Set the function and add all the kwargs as parameters.
        super().__init__(self._rvs, **kwargs)

    def _rvs(self):
        """A place holder function to use for object naming."""
        pass

    def set_seed(self, new_seed):
        """Update the random number generator's seed to a given value.

        Parameters
        ----------
        new_seed : int
            The given seed
        """
        self._rng = np.random.default_rng(seed=new_seed)

    def _create_and_sample(self, args, rng):
        """Create the distribution function and sample it. This is only
        needed if our distribution is in the form of a class that must
        be instantiated with different parameters each sampling run.

        Parameters
        ----------
        args : dict
            A dictionary mapping argument name to individual values.
        rng : numpy.random._generator.Generator
            The random number generator to use.

        Returns
        -------
        sample : float
            The result of sampling the function.
        """
        dist = self._dist(**args)
        sample = NumericalInversePolynomial(dist).rvs(1, rng)[0]
        return sample

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Execute the wrapped function.

        The input arguments are taken from the current graph_state and the outputs
        are written to graph_state.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.
        """
        rng = rng_info if rng_info is not None else self._rng

        if self._inv_poly is not None:
            # Batch sample all the results.
            num_samples = None if graph_state.num_samples == 1 else graph_state.num_samples
            results = self._inv_poly.rvs(num_samples, rng)
        else:
            # This is a class so we will need to create a new distribution object
            # for each sample (with a single instance of the input parameters).
            args = self._build_inputs(graph_state, **kwargs)

            if graph_state.num_samples == 1:
                dist = self._dist(**args)
                results = NumericalInversePolynomial(dist).rvs(1, rng)[0]
            else:
                # Transpose the dict of arrays to a list of dicts.
                arg_list = transpose_dict_of_list(args, graph_state.num_samples)
                results = self._vect_sample(arg_list, rng)

        # Save and return the results.
        self._save_results(results, graph_state)
        return results


class PDFFunctionWrapper:
    """A class that just wraps a given PDF function.

    Attributes
    ----------
    _pdf : function
        The PDF function.
    """

    def __init__(self, func):
        self._pdf = func
        self.pdf = self._pdf


class SamplePDF(NumericalInversePolynomialFunc):
    """A node for sampling from a given PDF function.

    Parameters
    ----------
    dist : function, class, or object
        The pdf function from which to sample or a class/object with that function.
    """

    def __init__(self, dist, **kwargs):
        if hasattr(dist, "pdf"):
            self.dist_obj = dist
        elif callable(dist):
            self.dist_obj = PDFFunctionWrapper(dist)
        else:
            raise ValueError("No pdf function detected.")
        super().__init__(self.dist_obj, **kwargs)


class LogPDFFunctionWrapper:
    """A class that just wraps a given Log PDF function.

    Attributes
    ----------
    _log_pdf : function
        The log PDF function.
    """

    def __init__(self, func):
        self._log_pdf = func
        self.logpdf = self._log_pdf


class SampleLogPDF(NumericalInversePolynomialFunc):
    """A node for sampling from a given Log PDF function.

    Parameters
    ----------
    dist : function, class, or object
        The pdf function from which to sample or a class/object with that function.
    """

    def __init__(self, dist, **kwargs):
        if hasattr(dist, "logpdf"):
            self.dist_obj = dist
        elif callable(dist):
            self.dist_obj = LogPDFFunctionWrapper(dist)
        else:
            raise ValueError("No logpdf function detected.")
        super().__init__(self.dist_obj, **kwargs)



"""Multiple object models wrap multiple BasePhysicalModels, allowing the
user to define such operations as additive models, where each object
contributes to the total flux density, or random object models, where
only one object is selected at random for each flux calculation.
"""

import numpy as np

from lightcurvelynx.graph_state import GraphState
from lightcurvelynx.math_nodes.given_sampler import GivenValueSampler
from lightcurvelynx.models.physical_model import BandfluxModel, BasePhysicalModel, SEDModel


class MultiObjectModel(SEDModel):
    """A MultiObjectModel wraps multiple BasePhysicalModels (including BandfluxModels).

    All rest frame effects are applied to each object, allowing different redshifts
    for each object (for unresolved objects).  The observer frame effects are applied
    to the weighted sum of the objects.

    While this model supports both BandfluxModels and SED, it inherits from SEDModel
    to pick up some of the helper functions.

    Note: Each object may have its own sampled (RA, dec) position, which are not
    required to align.

    Attributes
    ----------
    objects : list
        A list of BasePhysicalModel objects to use in the flux calculation.
    num_objects : int
        The number of objects in the model.
    _is_bandflux : list
        A list of Booleans indicating whether each model is a BandfluxModel.
    _any_bandflux : bool
        True if any of the models are bandflux models.

    Parameters
    ----------
    objects : list
        A list of BasePhysicalModel objects to use in the flux calculation.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(
        self,
        objects,
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Check that all objects are BasePhysicalModel objects and mark whether they are BandfluxModels.
        self._is_bandflux = [False] * len(objects)
        self._any_bandflux = False
        for idx, object in enumerate(objects):
            if isinstance(object, BandfluxModel):
                self._is_bandflux[idx] = True
                self._any_bandflux = True
            elif not isinstance(object, BasePhysicalModel):
                raise ValueError("All objects must be BasePhysicalModel objects.")

        self.objects = objects
        self.num_objects = len(objects)

    def set_graph_positions(self, seen_nodes=None):
        """Force an update of the graph structure (numbering of each node).

        Parameters
        ----------
        seen_nodes : set, optional
            A set of nodes that have already been processed to prevent infinite loops.
            Caller should not set.
        """
        if seen_nodes is None:
            seen_nodes = set()

        # Set the graph positions for this node and each object.
        super().set_graph_positions(seen_nodes=seen_nodes)
        for object in self.objects:
            object.set_graph_positions(seen_nodes=seen_nodes)

    def set_apply_redshift(self, apply_redshift):
        """Toggles the apply_redshift setting.

        Parameters
        ----------
        apply_redshift : bool
            The new value for apply_redshift.
        """
        for idx, object in enumerate(self.objects):
            if self._is_bandflux[idx]:
                object.set_apply_redshift(apply_redshift)

    def add_effect(self, effect):
        """Add an effect to each of the submodels.

        Parameters
        ----------
        effect : Effect
            The effect to add to the model.
        """
        if effect.rest_frame:
            # The rest frame effects are applied to each object.
            for object in self.objects:
                object.add_effect(effect)
        else:
            # Add observer frame effects to this model.
            # Add any effect parameters that are not already in the model.
            for param_name, setter in effect.parameters.items():
                if param_name not in self.setters:
                    self.add_parameter(param_name, setter, allow_gradient=False)
            self.obs_frame_effects.append(effect)

    def sample_parameters(self, given_args=None, num_samples=1, rng_info=None):
        """Sample the model's underlying parameters if they are provided by a function
        or ParameterizedModel.

        Parameters
        ----------
        given_args : dict, optional
            A dictionary representing the given arguments for this sample run.
            This can be used as the JAX PyTree for differentiation.
        num_samples : int
            A count of the number of samples to compute.
            Default: 1
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            All the keyword arguments, including the values needed to sample
            parameters.

        Returns
        -------
        graph_state : GraphState
            An object mapping graph parameters to their values.
        """
        # If the graph has not been sampled ever, update the node positions for every node.
        if self.node_pos is None:
            self.set_graph_positions()

        # We use the same seen_nodes for all sampling calls so each node
        # is sampled at most one time regardless of link structure.
        graph_state = GraphState(num_samples)
        if given_args is not None:
            graph_state.update(given_args, all_fixed=True)

        seen_nodes = {}
        for object in self.objects:
            object._sample_helper(graph_state, seen_nodes, rng_info=rng_info)
        self._sample_helper(graph_state, seen_nodes, rng_info=rng_info)

        return graph_state

    def _evaluate_single(self, times, wavelengths, state, rng_info=None, **kwargs):
        """Evaluate the model and apply the effects for a single, given graph state.
        This function applies redshift, computes the flux density for the object,
        applies rest frames effects, performs the redshift correction (if needed),
        and applies the observer frame effects.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray
            A length N array of wavelengths (in angstroms).
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            All the other keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        raise NotImplementedError

    def _evaluate_bandfluxes_single(self, passband_group, times, filters, state, rng_info=None) -> np.ndarray:
        """Get the band fluxes for a given PassbandGroup and a single, given graph state.

        Parameters
        ----------
        passband_group : PassbandGroup
            The passband group to use.
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filters : numpy.ndarray
            A length T array of filter names.
        state : GraphState
            An object mapping graph parameters to their values.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.

        Returns
        -------
        bandfluxes : numpy.ndarray
            A length T array of band fluxes for this sample.
        """
        raise NotImplementedError


class AdditiveMultiObjectModel(MultiObjectModel):
    """An AdditiveMultiObjectModel computes the flux from multiple overlapping objects,
    including (host galaxy and source pairs) or unresolved sources.

    All rest frame effects are applied to each model, allowing different redshifts
    for each model (for unresolved objects).  The observer frame effects are applied
    to the weighted sum of the models.

    Note: Each model may have its own sampled (RA, dec) position, which are not
    required to align.

    Attributes
    ----------
    objects : list
        A list of BasePhysicalModel objects to use in the flux calculation.
    weights : numpy.ndarray, optional
        A length N array of weights to apply to each object. If None, all objects
        will be weighted equally.
    num_objects : int
        The number of objects in the model.

    Parameters
    ----------
    objects : list
        A list of BasePhysicalModel objects to use in the flux calculation.
    weights : numpy.ndarray, optional
        A length N array of weights to apply to each object. If None, all objects
        will be weighted equally.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(
        self,
        objects,
        weights=None,
        **kwargs,
    ):
        super().__init__(objects, **kwargs)

        if weights is None:
            self.weights = np.full(len(objects), 1.0)
        elif len(weights) != len(objects):
            raise ValueError("Length of weights must match length of objects.")
        else:
            self.weights = weights

    def minwave(self, graph_state=None):
        """Get the minimum wavelength of the model. For additive models, this is
        a list of minimums for each object.

        Note
        ----
        Wavelength extrapolation is handled by each object. So the actual wavelength's
        can be evaluated outside the range of each object.

        Parameters
        ----------
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. If provided,
            the function will use the graph state to compute the minimum wavelength.

        Returns
        -------
        minwave : list of float or None
            The minimum wavelength of the each object (in angstroms) or None
        """
        return [object.minwave(graph_state=graph_state) for object in self.objects]

    def maxwave(self, graph_state=None):
        """Get the maximum wavelength of the model. For additive models, this is
        a list of maximums for each object.

        Note
        ----
        Wavelength extrapolation is handled by each object. So the actual wavelength's
        can be evaluated outside the range of each object.

        Parameters
        ----------
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. If provided,
            the function will use the graph state to compute the maximum wavelength.

        Returns
        -------
        maxwave : list of float or None
            The maximum wavelength of the each object (in angstroms) or None
        """
        return [object.maxwave(graph_state=graph_state) for object in self.objects]

    def _evaluate_single(self, times, wavelengths, state, rng_info=None, **kwargs):
        """Evaluate the model and apply the effects for a single, given graph state.
        This function applies redshift, computes the flux density for the object,
        applies rest frames effects, performs the redshift correction (if needed),
        and applies the observer frame effects.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray
            A length N array of wavelengths (in angstroms).
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            All the other keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        # Check that all models can compute SEDs.
        if self._any_bandflux:
            raise TypeError(
                "AdditiveMultiObjectModel contains at least one BandfluxModel, "
                "which does not support the evaluation of SEDs."
            )

        # Compute the weighted sum of contributions from each object. Since we use each
        # object's _evaluate_single function, the rest frame effects are applied
        # correctly for each object and wavelength extrapolation is handled by each object
        # (allowing them to have different wavelength ranges).
        flux_density = np.zeros((len(times), len(wavelengths)))
        for object, weight in zip(self.objects, self.weights, strict=False):
            flux_density += weight * object._evaluate_single(
                times,
                wavelengths,
                state,
                rng_info=rng_info,
                **kwargs,
            )

        # Apply the observer frame effects on the weighted sum of objects.
        params = self.get_local_params(state)
        for effect in self.obs_frame_effects:
            flux_density = effect.apply(
                flux_density,
                times=times,
                wavelengths=wavelengths,
                rng_info=rng_info,
                **params,
            )

        return flux_density

    def _evaluate_bandfluxes_single(self, passband_group, times, filters, state, rng_info=None) -> np.ndarray:
        """Get the band fluxes for a given PassbandGroup and a single, given graph state.

        Parameters
        ----------
        passband_group : PassbandGroup
            The passband group to use.
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filters : numpy.ndarray
            A length T array of filter names.
        state : GraphState
            An object mapping graph parameters to their values.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.

        Returns
        -------
        bandfluxes : numpy.ndarray
            A length T array of band fluxes for this sample.
        """
        # Compute the band fluxes as the weighted sum of all bandfluxes. All models will have rest frame
        # effects applied prior to summing. In the case of full SED models, the effects will be applied
        # to the entire SED before integrating with the filters to compute the band fluxes.
        bandfluxes = np.zeros(len(times))
        for idx, object in enumerate(self.objects):
            object_fluxes = object._evaluate_bandfluxes_single(
                passband_group,
                times,
                filters,
                state,
                rng_info=rng_info,
            )
            bandfluxes += self.weights[idx] * object_fluxes

        # Apply any common rest frame effects to the total bandflux.  We need to use the effects'
        # apply_bandflux() function since we no longer have SEDs.
        params = self.get_local_params(state)
        for effect in self.obs_frame_effects:
            bandfluxes = effect.apply_bandflux(
                bandfluxes,
                times=times,
                filters=filters,
                rng_info=rng_info,
                **params,
            )

        return bandfluxes


class RandomMultiObjectModel(MultiObjectModel):
    """A RandomMultiObjectModel selects one of its objects at random and
    computes the flux from that object.

    Attributes
    ----------
    object_map : dict
        A dictionary mapping each object name (or index) to a BasePhysicalModel object.
    num_objects : int
        The number of objects in the model.

    Parameters
    ----------
    objects : list
        A list of BasePhysicalModel objects to use in the flux calculation.
    weights : numpy.ndarray, optional
        A length N array indicating the relative weight from which to select
        a object at random. If None, all objects will be weighted equally.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(
        self,
        objects,
        weights=None,
        object_names=None,
        **kwargs,
    ):
        super().__init__(objects, **kwargs)

        # Create a parameter to indicate which object was selected.
        object_names = object_names or [src.node_string for src in objects]
        self.object_map = {name: src for name, src in zip(object_names, objects, strict=False)}
        self._sampler_node = GivenValueSampler(object_names, weights=weights)
        self.add_parameter("selected_object", value=self._sampler_node, allow_gradient=False)

    def minwave(self, graph_state=None):
        """Get the minimum wavelength of the model.

        Parameters
        ----------
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. If provided,
            the function will use the graph state to compute the minimum wavelength.

        Returns
        -------
        minwave : float or None
            The minimum wavelength of the model (in angstroms) or None
            if the model does not have a defined minimum wavelength.
        """
        idx = self.get_param(graph_state, "selected_object")
        return self.objects[idx].minwave(graph_state=graph_state)

    def maxwave(self, graph_state=None):
        """Get the maximum wavelength of the model.

        Parameters
        ----------
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. If provided,
            the function will use the graph state to compute the maximum wavelength.

        Returns
        -------
        maxwave : float or None
            The maximum wavelength of the model (in angstroms) or None
            if the model does not have a defined maximum wavelength.
        """
        idx = self.get_param(graph_state, "selected_object")
        return self.objects[idx].maxwave(graph_state=graph_state)

    def _evaluate_single(self, times, wavelengths, state, rng_info=None, **kwargs):
        """Evaluate the model and apply the effects for a single, given graph state.
        This function applies redshift, computes the flux density for the object,
        applies rest frames effects, performs the redshift correction (if needed),
        and applies the observer frame effects.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray
            A length N array of wavelengths (in angstroms).
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            All the other keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        # Check that all models can compute SEDs.
        if self._any_bandflux:
            raise TypeError(
                "RandomMultiObjectModel contains at least one BandfluxModel, "
                "which does not support the evaluation of SEDs."
            )

        # Use the model selected by the sampler node to compute the flux density.
        model_name = self.get_param(state, "selected_object")
        flux_density = self.object_map[model_name]._evaluate_single(
            times,
            wavelengths,
            state,
            rng_info=rng_info,
            **kwargs,
        )

        # Apply the observer frame effects on the selected object.
        params = self.get_local_params(state)
        for effect in self.obs_frame_effects:
            flux_density = effect.apply(
                flux_density,
                times=times,
                wavelengths=wavelengths,
                rng_info=rng_info,
                **params,
            )

        return flux_density

    def _evaluate_bandfluxes_single(self, passband_group, times, filters, state, rng_info=None) -> np.ndarray:
        """Get the band fluxes for a given PassbandGroup and a single, given graph state.

        Parameters
        ----------
        passband_group : PassbandGroup
            The passband group to use.
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filters : numpy.ndarray
            A length T array of filter names.
        state : GraphState
            An object mapping graph parameters to their values.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.

        Returns
        -------
        bandfluxes : numpy.ndarray
            A length T array of band fluxes for this sample.
        """
        # Use the model selected by the sampler node to compute the flux density.
        model_name = self.get_param(state, "selected_object")
        bandfluxes = self.object_map[model_name]._evaluate_bandfluxes_single(
            passband_group,
            times,
            filters,
            state,
            rng_info=rng_info,
        )

        # Apply the observer frame effects on the selected object.
        params = self.get_local_params(state)
        for effect in self.obs_frame_effects:
            bandfluxes = effect.apply_bandflux(
                bandfluxes,
                times=times,
                filters=filters,
                rng_info=rng_info,
                **params,
            )
        return bandfluxes



"""A collection of toy models that are primarily used for testing."""

import numpy as np

from lightcurvelynx.models.physical_model import SEDModel


class ConstantSEDModel(SEDModel):
    """A model with a constant SED over both wavelength and time.

    Parameterized values include:
      * brightness - The inherent brightness
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - No effect for static model. [from BasePhysicalModel]

    Parameters
    ----------
    brightness : float
        The inherent brightness
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, brightness, **kwargs):
        super().__init__(**kwargs)
        self.add_parameter("brightness", brightness, **kwargs)

    def compute_sed(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
            Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        params = self.get_local_params(graph_state)
        return np.full((len(times), len(wavelengths)), params["brightness"])


class StepModel(ConstantSEDModel):
    """A static model that is on for a fixed amount of time.

    Parameterized values include:
      * brightness - The inherent brightness
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - The time the step function starts, in MJD.
      * t1- The time the step function ends, in MJD.

    Parameters
    ----------
    brightness : float
        The inherent brightness
    t0 : float
        The time the step function starts, in MJD.
    t1 : float
        The time the step function ends, in MJD.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, brightness, t1, **kwargs):
        super().__init__(brightness, **kwargs)

        # t0 is added in the PhysicalModel constructor.
        self.add_parameter("t1", t1, **kwargs)

    def compute_sed(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
           Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        flux_density = np.zeros((len(times), len(wavelengths)))
        params = self.get_local_params(graph_state)

        time_mask = (times >= params["t0"]) & (times <= params["t1"])
        flux_density[time_mask] = params["brightness"]
        return flux_density


class SinWaveModel(SEDModel):
    """A model that emits a sine wave.

    flux = brightness + amplitude * sin(2 * pi * frequency * (time - t0))

    Parameterized values include:
      * brightness - The inherent brightness
      * frequency - The frequence of the sine wave.
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - The start of the sine wave's period. [from BasePhysicalModel]

    Parameters
    ----------
    brightness : float
        The inherent brightness.
        Default: 0.0
    amplitude : float
        The amplitude of the sine wave.
        Default: 0.0
    frequency : float
        The frequency of the sine wave.
        Default: 1.0
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, *, brightness=0.0, amplitude=0.0, frequency=1.0, **kwargs):
        super().__init__(**kwargs)
        self.add_parameter("brightness", brightness, **kwargs)
        self.add_parameter("amplitude", amplitude, **kwargs)
        self.add_parameter("frequency", frequency, **kwargs)

    def compute_sed(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
            Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        params = self.get_local_params(graph_state)
        phases = 2.0 * np.pi * params["frequency"] * (times - params["t0"])
        single_wave = params["brightness"] + params["amplitude"] * np.sin(phases)
        return np.tile(single_wave[:, np.newaxis], (1, len(wavelengths)))


class LinearWavelengthModel(SEDModel):
    """A model that emits flux as a linear function of wavelength
    (that is constant over time): f(t, w) = scale * w + base.

    Includes optional minimum and maximum wavelength bounds to test
    extrapolation.

    Parameterized values include:
      * linear_base - The base brightness in nJy.
      * linear_scale - The slope of the linear function in nJy/Angstrom.
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - No effect for static model. [from BasePhysicalModel]

    Attributes
    ----------
    min_wave : float or None
        The minimum wavelength of the model (in angstroms). Or None if there
        is no minimum wavelength.
    max_wave : float or None
        The maximum wavelength of the model (in angstroms). Or None if there
        is no maximum wavelength.

    Parameters
    ----------
    linear_base : parameter
        The base brightness in nJy.
    linear_scale : parameter
        The slope of the linear function in nJy/Angstrom.
    min_wave : float or None
        The minimum wavelength of the model (in angstroms). Or None if there
        is no minimum wavelength.
    max_wave : float or None
        The maximum wavelength of the model (in angstroms). Or None if there
        is no maximum wavelength.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, linear_base, linear_scale, min_wave=None, max_wave=None, **kwargs):
        super().__init__(**kwargs)
        self.add_parameter("linear_base", linear_base, **kwargs)
        self.add_parameter("linear_scale", linear_scale, **kwargs)
        self.min_wave = min_wave
        self.max_wave = max_wave

    def minwave(self, graph_state=None):
        """Get the minimum wavelength of the model.

        Parameters
        ----------
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. Not used
            for this model.

        Returns
        -------
        minwave : float or None
            The minimum wavelength of the model (in angstroms) or None
            if the model does not have a defined minimum wavelength.
        """
        return self.min_wave

    def maxwave(self, graph_state=None):
        """Get the maximum wavelength of the model.

        Parameters
        ----------
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. Not used
            for this model.

        Returns
        -------
        maxwave : float or None
            The maximum wavelength of the model (in angstroms) or None
            if the model does not have a defined maximum wavelength.
        """
        return self.max_wave

    def compute_sed(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
            Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        params = self.get_local_params(graph_state)
        single_wave = params["linear_base"] + params["linear_scale"] * wavelengths
        return np.tile(single_wave[np.newaxis, :], (len(times), 1))
